import argparse
import csv
import genanki
import glob
import os
import pandas as pd
import random
import re
import sqlite3
import sys
import pprint

from dataclasses import dataclass
from pathlib import Path
from pybtex.database import parse_file

from ..io import data_export
from ..utils import argparser as my_argparse

# -----------
# latex stuff
# -----------

latex_packages = r"""\usepackage[T1]{fontenc}
\usepackage[utf8]{inputenc}
\usepackage[english, italian]{babel}
\usepackage{minitoc}
\usepackage{mypkg}
\usepackage{tikz}
\usetikzlibrary{arrows,snakes,backgrounds,calc}
\usepackage{wrapfig}
\usepackage{subfig}
\usepackage{rotating}
\usepackage{cancel}
\usepackage{hyperref}
"""

latex_preamble = (
    r"\documentclass[avery5371, grid]{flashcards}"
    + latex_packages
    + r"\begin{document}"
)
latex_ending = r"\end{document}"

# ----------
# anki stuff
# ----------

# Aggiunta di preamboli latex custom

# To use the package with Anki, click "Add" in the main window, then
# click the note type selection button. Click the "Manage" button,
# then select the note type you plan to use and click "Options". The
# LaTeX header and footer are shown. The header will look something
# like:

# \documentclass[12pt]{article}
# \special{papersize=3in,5in}
# \usepackage{amssymb,amsmath}
# \pagestyle{empty}
# \setlength{\parindent}{0in}
# \begin{document}

# To use chemtex, you’d add the usepackage line in the earlier example, so it looks like:

# \documentclass[12pt]{article}
# \special{papersize=3in,5in}
# \usepackage{amssymb,amsmath}
# \usepackage{chemtex}
# \pagestyle{empty}
# \setlength{\parindent}{0in}
# \begin{document}

# After that, you should be able to include lines like the following in your Anki cards:

# Per quanto riguarda le API di genanki per creare il modello di carta fare riferimento a

# https://github.com/kerrickstaley/genanki/blob/master/genanki/model.py
anki_latex_preamble = (
    "\\documentclass[12pt]{article}\n" + latex_packages + "\\begin{document}"
)
model_id = 1607392319
deck_id = 2059400110
card_model_name = 'model_with_my_latex_packages'

card_model = genanki.Model(
    model_id,
    card_model_name,
    fields=[
        {'name': 'Domanda'},
        {'name': 'Risposta'},
    ],
    templates=[
        {
            'name': 'Card 1',
            'qfmt': '{{Domanda}}',
            'afmt': '{{Risposta}}',
        },
    ],
    latex_pre=anki_latex_preamble,
)


@dataclass
class Card:
    s1: str = ""
    s2: str = ""

    def to_tex(self) -> None:
        return (
            (r"\begin{flashcard}{%s}" % self.s1)
            + "   "
            + self.s2
            + "   "
            + r"\end{flashcard}"
        )

    def to_csv(self) -> tuple:
        return (self.s1, self.s2)

    def to_anki(self, add_latex_tags):
        s1 = (
            "[latex] {} [/latex]".format(self.s1)
            if add_latex_tags
            else self.s1
        )
        s2 = (
            "[latex] {} [/latex]".format(self.s2)
            if add_latex_tags
            else self.s2
        )
        return (s1, s2)


class Flashcards(object):
    def __init__(
        self,
        path: str | Path,
        latex_envirs: list[str] = [
            "thm",
            "cor",
            "lem",
            "prop",
            "proof",
            "defn",
            "es",
            "rmk",
        ],
    ):
        # initialization: flashcards list (lista di tuple) e regex per
        # gli env latex
        self.__fc = []
        paste = "|".join(latex_envirs)
        fmt = (paste, paste)
        self.__env_re = re.compile(
            r"\\begin{(%s)}(\[.+?\])?(.+?)\\end{(%s)}" % fmt
        )
        # load data from files given in the path
        path = Path(path)
        if path.is_dir():
            files = [
                f
                for f in path.iterdir()
                if (
                    (f.suffix in ('.tex', '.Rnw', '.csv'))
                    and f.name != "_region_.tex"
                )
            ]
        else:
            files = [path]
        for f in files:
            if f.suffix in ('.tex', '.Rnw'):
                self.add_from_tex(f)
            elif f.suffix == '.csv':
                self.add_from_csv(f)

    def add_from_csv(self, path: str | Path) -> None:
        path = Path(path)
        with path.open() as csvfile:
            reader = csv.reader(csvfile)
            for row in reader:
                self.__fc.append(Card(row[0], row[1]))

    def add_from_tex(self, path: str | Path) -> None:
        path = Path(path)
        # import as list of tuples
        with path.open() as t:
            tmp = t.readlines()
            # rm commented stuff and join
            tmp = [l for l in tmp if not l.lstrip().startswith("%")]
            tmp = " ".join(tmp)
            # remove newline and duplicate spaces
            tmp = tmp.replace("\n", "")
            tmp = re.sub(r"\\label{.+?}", "", tmp)
            tmp = re.sub("\s+", " ", tmp)
            matches = self.__env_re.findall(tmp)
            for match in matches:
                if len(match) == 3:  # no [] for the environment, only content
                    side1 = "[{0}]".format(match[0])  # name of the environment
                    content = match[1]
                elif len(match) == 4:  # both [] and environment content
                    rm_paren = match[1].replace("[", "").replace("]", "")
                    side1 = "[{0}]".format(match[0]) + " " + rm_paren
                    content = match[2]
                self.__fc.append(Card(side1, content))

    def to_csv(self, path: str | Path) -> None:
        '''Export to a csv'''
        if path is None:
            path = Path("/tmp/flashcards.csv")
        else:
            path = Path(path)
        with path.open(mode='w') as f:
            dataset = csv.writer(
                f, delimiter=';', quotechar='"', quoting=csv.QUOTE_NONNUMERIC
            )
            for card in self.__fc:
                dataset.writerow(card.to_csv())
        print("All done, exported to: " + str(path))

    def to_xlsx(self, path: str | Path) -> None:
        '''Export to a xlsx'''
        if path is None:
            path = Path("/tmp/flashcards.xlsx")
        else:
            path = Path(path)
        data = []
        for card in self.__fc:
            side1, side2 = card.to_csv()
            data.append([side1, side2])
        df = pd.DataFrame(data, columns=["side1", "side2"])
        data_export({"flashcards": df}, path=path)
        print("All done, exported to: " + str(path))

    def to_tex(self, path: str | Path) -> None:
        if path is None:
            path = Path("/tmp/flashcards.tex")
        else:
            path = Path(path)
        with path.open(mode="w") as f:
            print(latex_preamble, file=f)
            for card in self.__fc:
                print(card.to_tex(), "\n", file=f)
            print(latex_ending, file=f)
        print("All done, now run:\n\t pdflatex " + str(path))

    def to_anki(
        self,
        path: str | Path | None,
        deck_name: str | None,
        add_latex_tags: bool,
    ):
        """Export to anki"""
        if deck_name is None:
            deck_name = "test"
        if path is None:
            path = Path("/tmp/flashcards.apkg")
        else:
            path = Path(path)
        deck = genanki.Deck(deck_id, deck_name)
        for card in self.__fc:
            elem = card.to_anki(add_latex_tags=add_latex_tags)
            note = genanki.Note(model=card_model, fields=[elem[0], elem[1]])
            deck.add_note(note)
        genanki.Package(deck).write_to_file(path)

    def export(self, outfile: str | Path, infile_ext: str):
        outfile = Path(outfile)
        ext = outfile.suffix
        if ext == '.csv':
            self.to_csv(path=outfile)
        if ext == '.xlsx':
            self.to_xlsx(path=outfile)
        if ext == '.tex':
            self.to_tex(path=outfile)
        if ext == '.apkg':
            add_latex_tags = True if infile_ext == '.tex' else False
            self.to_anki(
                path=outfile,
                deck_name=outfile.stem,
                add_latex_tags=add_latex_tags,
            )


def flashcards():
    parser = argparse.ArgumentParser()
    parser.add_argument("infile")
    parser.add_argument("outfile")
    args = parser.parse_args()
    infile = Path(args.infile).resolve()
    infile_ext = infile.suffix
    outfile = Path(args.outfile).resolve()
    if not infile.exists():
        msg = str(infile) + " does not exists."
        raise FileNotFoundError(msg)
    fc = Flashcards(infile)
    fc.export(outfile, infile_ext)


tex_default_preamble = '''
\\documentclass[a4paper]{article}
\\usepackage[T1]{fontenc}
\\usepackage[utf8]{inputenc}
\\usepackage[english, italian]{babel}
\\usepackage{amsmath}
\\usepackage{amsfonts}
\\usepackage{amssymb}
\\usepackage{mathrsfs}
\\usepackage{amsthm}
\\usepackage{tikz}
\\usepackage{mypkg}
\\usepackage{hyperref}
'''


def make_unique_list(x):
    if x is None:
        return []
    if not isinstance(x, list):
        x = [x]
    x = list(set(x))
    return x


class Database(object):
    '''Class to create a database (sqlite3) file'''

    __exercise_re = re.compile(r'\\begin{exercise}.+?\\end{exercise}')
    __id_re = re.compile(r'\\id{(.+?)}')
    __page_re = re.compile(r'\\page{(.+?)}')
    __source_re = re.compile(r'\\source{(.+?)}')
    __topic_re = re.compile(r'\\topic{(.+?)}')
    __question_re = re.compile(r'\\begin{question}(.+?)\\end{question}')
    __hint_re = re.compile(r'\\begin{hint}(.+?)\\end{hint}')
    __solution_re = re.compile(r'\\begin{solution}(.+?)\\end{solution}')

    __sql_create_ex_table = '''CREATE TABLE exercises
    (id text, page int, source text, topic text, question text, 
    hint text, solution text)'''
    __sql_create_biblio_table = '''CREATE TABLE sources
    (key text, title text, author text, subject text)'''
    __sql_insert_ex_table = '''INSERT INTO exercises VALUES
    (?, ?, ?, ?, ?, ?, ?)'''
    __sql_insert_biblio_table = '''INSERT INTO sources VALUES
    (?, ?, ?, ?)'''

    def __init__(self):
        # list like this [(), ()] which are easy to handle with sqlite3
        self.exercises_list = []
        self.biblio_list = []
        self.__parsed = set()  # already parsed paths

    # ------------
    # main methods
    # ------------
    def feed(self, paths=None, paths_f=None, biblio_dirs=None):
        '''Read exercises from comma separated paths (directory/files)
        and/from files of paths'''
        if paths is None:
            paths = []
        if paths_f:  # se non è None aggiungi i path inseriti nei file
            paths_f_paths = []
            for pf in paths_f:
                pf = Path(pf).expanduser()
                if pf.is_file():
                    with open(pf, 'r') as f:
                        # keep it if not comment or blank line
                        paths_f_paths += [
                            l
                            for l in f.read().splitlines()
                            if ((not l.startswith("#")) and (l.strip() != ""))
                        ]
            paths += paths_f_paths
        if biblio_dirs is None:
            biblio_dirs = ["~/src/other/exercises/biblio"]
        paths += biblio_dirs
        if not paths:
            msg = "Cosa strana che non vi sian elementi qui ne da paths ne da list"
            raise Exception(msg)
        # now start parsing recursively: if file, parse it; if dir: go recursive
        for p in paths:
            p = Path(p).expanduser()
            # controllo che non sia gia stato parsato per evitare cicli infiniti
            if p in self.__parsed:
                print(p, 'è già stato parsato, lo salto.')
                continue
            if p.is_dir():
                self.__parse_dir(d=p)
                self.__parsed.add(p)
            elif p.is_file():
                self.__parse_file(f=p)
                self.__parsed.add(p)
            else:
                print(p, 'non è ne una directory ne un file, lo salto.')
        print("Processed files:")
        pprint.pprint(self.__parsed)
        print("Extracted exercises:", len(self.exercises_list))
        return self

    def __parse_dir(self, d):
        # effettua lo stesso ciclo di feed sul contenuto della singola directory
        # if file: parse it
        # if dir: go recursive on each entry
        # else do nothing
        for p in d.iterdir():
            if p in self.__parsed:
                print(p, 'è già stato parsato, lo salto.')
                continue
            if p.is_dir():
                self.__parse_dir(d=p)
                self.__parsed.add(p)
            elif p.is_file():
                self.__parse_file(f=p)
                self.__parsed.add(p)
            else:
                print(p, 'non è ne una directory ne un file, lo salto.')

    def __parse_file(self, f):
        # parsa un singolo file in base all'estensione
        ext = f.suffix
        if ext in {".tex"}:
            self.__parse_tex(path=f)
        elif ext in {".bib"}:
            self.__parse_bib(path=f)
        else:
            print(f, 'non è un file .tex o .bib; ignoro.')

    def write(self, f=None):
        'Write exercises to a sqlite3 file'
        # erase old version
        f = os.path.expanduser(f)
        if os.path.isfile(f):
            os.remove(f)
        # create exercise table
        con = sqlite3.connect(f)
        c = con.cursor()
        c.execute(self.__sql_create_ex_table)
        c.execute(self.__sql_create_biblio_table)
        # fill exercise table
        c.executemany(self.__sql_insert_ex_table, self.exercises_list)
        c.executemany(self.__sql_insert_biblio_table, self.biblio_list)
        con.commit()
        con.close()
        return self

    # ----------------
    # helper functions
    # ----------------
    def __parse_tex(self, path=None):
        # reading
        with open(path, 'r') as f:
            data = f.read()
        data = data.replace('\n', '')
        data = re.sub('\s+', ' ', data)
        # parsing
        # 1) split by exercise
        exercises = self.__exercise_re.findall(data)
        # 2) extract and save
        for ex in exercises:
            self.exercises_list.append(self.__parse_exercise(ex=ex, f=path))

    def __get_regex_value(self, re, x):
        found = re.search(x)
        if found:
            res = found.group(1).strip()
            # TODO fix here
            # return(re.sub(' +', ' ', res))
            return res
        else:
            return None

    def __parse_exercise(self, ex=None, f=None):
        ID = self.__get_regex_value(self.__id_re, ex)
        page = self.__get_regex_value(self.__page_re, ex)
        source = self.__get_regex_value(self.__source_re, ex)
        topic = self.__get_regex_value(self.__topic_re, ex)
        question = self.__get_regex_value(self.__question_re, ex)
        hint = self.__get_regex_value(self.__hint_re, ex)
        solution = self.__get_regex_value(self.__solution_re, ex)
        try:
            page = int(page)
        except ValueError:
            page = None
            print('Invalid integer page in ' + str(f) + '. Setting it to None')
        return (ID, page, source, topic, question, hint, solution)

    def __parse_bib(self, path=None):
        bib_data = parse_file(path)
        for entry in bib_data.entries.values():
            # keep only those with a subject
            if 'subject' in entry.fields:
                cognomi = []
                for author in entry.persons["author"]:
                    cognomi.append(str(author.last_names[0]))
                self.biblio_list.append(
                    (
                        entry.key,
                        entry.fields['title'],
                        ", ".join(cognomi),
                        entry.fields['subject'],
                    )
                )


# -------------------------------------------------------------------


class Worksheet(object):
    '''Class to extract exercises from a database file'''

    def __init__(self):
        # list like this [(), ()] which are easy to handle with sqlite3
        self.exercises_list = []

    def select(self, files=None, select_where='', random=False, n=-1):
        'Do a sql select on a sqlite3 database and save rows in exercises_list'

        # avoid duplicated file names
        files = make_unique_list(files)

        # row selection
        if isinstance(select_where, str) and select_where != '':
            where = " and " + select_where
        else:
            where = ""

        # random order
        if isinstance(random, bool) and random is True:
            order_by = " order by random() "
        else:
            order_by = ""

        # setup limit
        if isinstance(n, int) and n >= 0:
            limit = " limit " + str(n)
        else:
            limit = ""

        sql = (
            '''select id, page, source, topic, question,
                        hint, solution, subject
                 from   exercises, sources
                 where  exercises.source = sources.key '''
            + where
            + order_by
            + limit
        )

        anydb = False
        for f in files:
            f = os.path.expanduser(f)
            if os.path.isfile(f):
                con = sqlite3.connect(f)
                c = con.cursor()
                try:
                    c.execute(sql)
                except sqlite3.OperationalError:
                    print(
                        "Errore nell'esecuzione di:\n ", "              ", sql
                    )
                self.exercises_list += c.fetchall()
                con.close()
                anydb = True
            else:
                print(f, ' does not exists. Skipping\n')
        if not anydb:
            msg = "No useful db files were provided"
            raise Exception(msg)
        return self

    def to_tex(
        self,
        tex='/tmp/worksheet.tex',
        preamble=tex_default_preamble,
        show_topic=True,
        show_hint=True,
    ):
        '''Export exercises_list in a .tex'''
        tex = os.path.expanduser(tex)
        doc = (
            [preamble, '''\\begin{document}''']
            + ['\\tableofcontents']
            + self.__format_exercises(
                show_topic=show_topic, show_hint=show_hint
            )
            + ['''\\end{document}''']
        )
        # here replace None with ''
        for i in range(0, len(doc)):
            if doc[i] is None:
                doc[i] = ""
        # and output to .tex file
        with open(tex, 'w') as f:
            f.write('\n'.join(doc))
        return self

    def __format_exercises(self, show_topic, show_hint):
        ex_section = []
        res_section = []

        column = {
            'id': 0,
            'page': 1,
            'source': 2,
            'topic': 3,
            'question': 4,
            'hint': 5,
            'solution': 6,
            'subject': 7,
        }

        index_list = list(range(0, len(self.exercises_list)))
        for i in index_list:
            ex_id = (self.exercises_list[i])[column['id']]
            page = (self.exercises_list[i])[column['page']]
            source = (self.exercises_list[i])[column['source']]
            topic = (self.exercises_list[i])[column['topic']]
            question = (self.exercises_list[i])[column['question']]
            hint = (self.exercises_list[i])[column['hint']]

            if show_hint and (hint is not None):
                hint = '\\emph{Suggerimento}: ' + hint
            else:
                hint = ''

            header_tail = (
                ex_id + ' pag.~' + str(page) + ' (' + str(source) + ')' + '}'
            )

            if show_topic and (topic is not None):
                header_es = '\\paragraph{' + topic + ': es.~' + header_tail
            else:
                header_es = '\\paragraph{Es.~' + header_tail

            header_sol = '\\paragraph{Sol.~es.~' + header_tail

            if question is not None or question == '':
                soluzione = (self.exercises_list[i])[column['solution']]
                ex_section += [header_es] + [question] + [hint]
                if soluzione is not None or soluzione == '':
                    res_section += [header_sol] + [soluzione]
        return (
            ['''\\section{Esercizi}''']
            + ex_section
            + ['''\\newpage''']
            + ['''\\section{Soluzioni}''']
            + res_section
        )


def exercises_db():
    opts = (
        # (param, help, default, type)
        # --paths
        (
            'paths',
            'str: comma separated list of exercise paths (source directories/files)',
            # '~/src/other/exercises',
            None,
            str,
        ),
        # --lists
        (
            'lists',
            'str: comma separated list of file having a lists of paths (dir/files) one per line',
            None,
            str,
        ),
        # --outfile
        (
            'outfile',
            'str:  sqlite3 db to save',
            None,
            # '~/.exercises.db',
            str,
        ),
    )

    args = my_argparse(opts)
    paths = args['paths']
    lists = args['lists']
    outfile = args['outfile']
    # se sono entrambi a none c'è qualcosa che non torna
    if (paths is None) and (lists is None):
        msg = "uno tra --paths e --lists deve essere specificato"
        raise Exception(msg)
    if outfile is None:
        msg = "Bisogna specificare  il file sqlite3 su cui esportare mediante --outfile"
        raise Exception(msg)
    # da qui in poi dovrebbe essere a posto
    if isinstance(paths, str):
        paths = paths.split(',')
    if isinstance(lists, str):
        lists = lists.split(',')
    ex = Database()
    # print(paths, lists)
    ex.feed(paths=paths, paths_f=lists).write(outfile)
    return 0


def exercises_ws():
    opts = (
        # (param, help, default, type)
        # --dbs
        (
            'dbs',
            'str: comma separated list of db produced by exercises_db',
            # '~/.exercises_dbs',
            None,
            str,
        ),
        # --select_where
        (
            'select_where',
            'char: sql where statement used to select rows from the database',
            '',
            str,
        ),
        # --random
        ('random', 'bool: random ordering? (default: False)', False, bool),
        # --n
        (
            'n',
            'int: n. of records (if negative - the default - take them all)',
            -1,  #
            int,
        ),
        # --show_topic
        (
            'show_topic',
            'bool: show exercise topic? (default: True)',
            True,
            bool,
        ),
        # --show_hint
        ('show_hint', 'bool: show exercise hint? (default: True)', True, bool),
        # --tex
        ('tex', 'str:  output tex file', '/tmp/worksheet.tex', str),
    )
    args = my_argparse(opts)
    dbs = args['dbs'].split(',')
    select_where = args['select_where']
    random = args['random']
    n = args['n']
    show_topic = args['show_topic']
    # print(show_topic)
    show_hint = args['show_hint']
    # print(show_hint)
    tex = args['tex']
    ws = Worksheet()
    ws.select(files=dbs, select_where=select_where, random=random, n=n)
    ws.to_tex(tex=tex, show_topic=show_topic, show_hint=show_hint)
    return 0



"""Miscellaneous utilities"""

import argparse as _argparse
import re as _re
import readline as _readline

from typing import Sequence as _Sequence
from .iter import unique as _unique

_readline.parse_and_bind('set editing-mode emacs')


def argparser(opts):
    '''Helper function for argument parsing.

    Example usage:
    --------------
    import pylbmisc as lb # third party
    # from ..utils import argparser # in scripts directory

    opts = (
        # (param, help, default, type)
        # --dirs
        (
            'dirs',
            'str: comma separated list of exercise source directories',
            '~/src/pypkg/exercises/db',
            str,
        ),
        # --lists
        (
            'lists',
            'str: comma separated list of file with lists of source dir',
            None,
            str,
        ),
        # --outfile
        ('outfile', 'str:  sqlite3 db to save', '~/.exercises.db', str),
    )

    args = lb.utils.argparser(opts)
    dirs = args['dirs']
    dirs = dirs.split(',')
    lists = args['lists']
    lists = lists.split(',')
    outfile = args['outfile']
    print({"dirs": dirs, "lists": lists, "outfile": outfile})
    return 0

    '''
    parser = _argparse.ArgumentParser()
    # defaults = {}
    for i in opts:
        optname = i[0]
        optdescription = i[1]
        optdefault = i[2]
        opttype = i[3]
        # create help string and add argument to parsing
        help_string = '{0} (default: {1})'.format(
            optdescription, str(optdefault)
        )
        parser.add_argument('--' + optname, help=help_string, type=str)
    # do parsing
    args = vars(parser.parse_args())  # vars to change to a dict
    # defaults settings and types management
    for i in opts:
        optname = i[0]
        optdescription = i[1]
        optdefault = i[2]
        opttype = i[3]
        # se il valore è a none in args impostalo al valore di default
        # specificato
        if args[optname] is None:
            args[optname] = optdefault
        # se il tipo è logico sostituisci un valore possibile true con
        # l'equivalente python
        if opttype == bool:
            # mv to character if not already (not if used optdefault)
            args[optname] = str(args[optname])
            true_values = (
                'true',
                'True',
                'TRUE',
                't',
                'T',
                '1',
                'y',
                'Y',
                'yes',
                'Yes',
                'YES',
            )
            if args[optname] in true_values:
                args[optname] = 'True'
            else:
                args[optname] = ''
        # converti il tipo a quello specificato, a meno che non sia None se no lascialo così
        if args[optname] is not None:
            args[optname] = opttype(args[optname])
    return args


def line_to_numbers(x: str) -> list[int]:
    '''transform a string of positive numbers "1 2-3, 4, 6-10" into a
    list [1,2,3,4,6,7,8,9,10]
    '''
    # replace comma with white chars
    x = x.replace(",", " ")
    # keep only digits, - and white spaces
    x = _re.sub(r'[^\d\- ]', '', x)
    # split by whitespaces
    spl = x.split(" ")
    # change ranges to proper
    expanded = []
    single_page_re = _re.compile("^[0-9]+$")
    pages_range_re = _re.compile("^([0-9]+)-([0-9]+)$")
    for i in range(len(spl)):
        # Check if the single element match one of the regular expression
        single_page = single_page_re.match(spl[i])
        pages_range = pages_range_re.match(spl[i])
        if single_page:
            # A) One single page: append it to results
            expanded.append(spl[i])
        elif pages_range:
            # B) Pages range: append a list of (expanded) pages to results
            first = int(pages_range.group(1))
            second = int(pages_range.group(2))
            # step is 1 if first is less than or equal to second or -1
            # otherwise
            step = 1 * int(first <= second) - 1 * int(first > second)
            if step == 1:
                second += 1
            elif step == -1:
                second -= 1
            else:
                # do nothing (ignore if they don't match)
                pass
            expanded_range = [str(val) for val in range(first, second, step)]
            expanded += expanded_range
        else:
            ValueError(
                str(spl[i])
                + "does not match a single page re nor a pages range re."
            )
    # coerce to integer expanded
    res: list[int] = [int(x) for x in expanded]
    return res


def menu(
    choices: _Sequence[str],
    title: str | None = None,
    multiple: bool = False,
    repeated: bool = False,
    strict: bool = True,
) -> _Sequence[str]:
    """
    CLI menu for user single/multiple choices

    Returns either:
    - a single choice
    - a list of selected choiches
    - None if nothing was choosed

    TODO: add a default options
    """
    available_ind = [i + 1 for i in range(len(choices))]
    avail_with_0 = [0] + available_ind
    the_menu = "\n".join(
        [str(i) + '. ' + str(c) for i, c in zip(available_ind, choices)]
    )
    if multiple:
        select_msg = "Selection (values as '1, 2-3, 6') or 0 to exit: "
    else:
        select_msg = "Selection (0 to exit): "
    if title:
        ascii_header(title)
    print(the_menu, '\n')
    ind = line_to_numbers(input(select_msg))
    # normalize to list (for single selections, for now)
    if not isinstance(ind, list):
        ind = list(ind)
    if strict:
        # continue asking for input until all index are between the selectable
        while not all([i in avail_with_0 for i in ind]):
            not_in = [i for i in ind if i not in avail_with_0]
            print("Not valid insertion: ", not_in, "\n")
            ind = line_to_numbers(input(select_msg))
            if not isinstance(ind, list):
                ind = list(ind)
    else:
        # keep only the input in avail_with_0
        allowed = [i for i in ind if i in avail_with_0]
        any_not_allowed = not all(allowed)
        if any_not_allowed:
            print(
                "Removed some values (not 0 or specified possibilities): ",
                list(set(ind) - set(allowed)),
                ".",
            )
            ind = allowed
    # make unique if not allowed repetitions
    if not repeated:
        ind = list(_unique(ind))
    # obtain the selection
    rval = [choices[i - 1] for i in ind if i != 0]
    # return always a list should simplify the code
    return rval


def ascii_header(x: str) -> None:
    '''
    Create an ascii header given a string as title.
    '''
    header = "=" * len(x)
    print(header)
    print(x)
    print(header, '\n')


def match_arg(arg, choices):
    """R's match.arg equivalent for programming function for interactive use

    >>> # questo ritorna errore perché matcha troppo
    >>> user_input = "foo"
    >>> a = match_arg(user_input, ["foobar", "foos", "asdomar"])

    >>> # questo è ok e viene espanso
    >>> user_input2 = "foob"
    >>> a = match_arg(user_input2, ["foobar", "foos", "asdomar"])
    >>> print(a)
    """
    res = [expanded for expanded in choices if expanded.startswith(arg)]
    choices_str = ", ".join(choices)
    l = len(res)
    if l == 0:
        msg = f"Parameter {arg} must be one of: {choices_str}"
        raise ValueError(msg)
    elif l > 1:
        msg = f"Parameter {arg} matches multiple choices from: {choices_str}"
        raise ValueError(msg)
    else:
        return res[0]



"""Some iterators remembering R"""

from typing import Any as _Any
from typing import Iterator as _Iterator


def unique(x) -> _Iterator[_Any]:
    '''
    return a stream of unique items contained in x
    '''
    seen = set()
    for item in x:
        if item in seen:
            pass
        else:
            seen.add(item)
            yield item


# minor diffs from 1.10 of python cookbook
def duplicated(x) -> _Iterator[bool]:
    '''
    return a stream of logical value (already seen element)
    '''
    seen = set()
    for item in x:
        if item in seen:
            yield True
        else:
            seen.add(item)
            yield False


if __name__ == '__main__':
    alist = [1, 2, 1, 3]
    print(list(duplicated(alist)))



"""Data management utilities for pandas Series/DataFrame"""

import functools as _functools
import inspect as _inspect
import numpy as _np
import pandas as _pd
import re as _re
import string as _string
import subprocess as _subprocess
import tempfile as _tempfile

from pprint import pprint as _pprint

# -------------------------------------------------------------------------
# regular expressions used
# -------------------------------------------------------------------------

# Searching by regex in data
# https://stackoverflow.com/questions/51170763 for general setup
# Mail: https://stackoverflow.com/questions/8022530/ for mail regex
_mail_re = _re.compile(r"[^@]+@[^@]+\.[^@]+")
_fc_re = _re.compile(
    r"[A-Za-z]{6}[0-9]{2}[A-Za-z]{1}[0-9]{2}[A-Za-z]{1}[0-9]{3}[A-Za-z]{1}"
)
_tel_re = _re.compile(r"(.+)?0[0-9]{1,3}[\. /\-]?[0-9]{6,7}")
_mobile_re = _re.compile(r"(.+)?3[0-9]{2}[\. /\-]?[0-9]{6,7}")

# -------------------------------------------------------------------------
# Utilities
# -------------------------------------------------------------------------

def view(df: _pd.DataFrame):
    """View a pd.DataFrame using LibreOffice."""
    tempfile = _tempfile.mkstemp(suffix=".xlsx")
    fname = tempfile[1]
    df.to_excel(fname)
    _subprocess.Popen(["libreoffice", fname])


def table2df(df: _pd.DataFrame):
    """Transform a pd.DataFrame representing a two-way table (es
    crosstable, correlation matrix, p.val matrix) in a
    pd.DataFrame with long format.
    """
    x = df.copy()
    x = x.stack().reset_index()
    return x.rename(columns={0: "x"})


def dump_unique_values(dfs, fpath = "data/uniq_"):
    """Save unique value of a (dict of) dataframe for inspection and monitor during time."""

    if not (isinstance(dfs, _pd.DataFrame) or isinstance(dfs, dict)):
        msg = "x deve essere un pd.DataFrame o un dict di pd.DataFrame"
        raise ValueError(msg)

    # normalize single dataframe
    if isinstance(dfs, _pd.DataFrame):
        dfs = {"df": dfs}

    for df_lab, df in dfs.items():
        outfile = fpath + df_lab + ".txt"
        with open(outfile, "w") as f:
            for col in df:
                # Header
                print(f"DataFrame: '{df_lab}', Column: '{col}', Dtype: {df[col].dtype}, Unique values:", file = f)
                # Dati: non sortati perché ci sono problemi se i dati sono metà numerici e meta stringa
                # _pprint(df[col].sort_values().unique().tolist(), stream = f, compact = True)
                _pprint(df[col].unique().tolist(), stream = f, compact = True)
                print(file = f)


# -------------------------------------------------------------------------
# pii_erase
# ------------------------------------------------------------------------


# Searching by column name
def _columns_match(df_columns, searched):
    """Search in columns names, both exact match and contains match"""
    # coerce single string to list of one string
    if isinstance(searched, str):
        searched = [searched]
    perfect_match = [c in set(searched) for c in df_columns]
    return perfect_match



is_email = _np.vectorize(lambda x: bool(_mail_re.match(x)))



def _is_string(x: _pd.Series):
    """
    Check if a Series is a string (including data with np.nan) differently from
    pandas.api.types.is_string_dtype

    >>> x = pd.Series(["a", np.nan])
    >>> pd.api.types.is_string_dtype(x)
    >>> _is_string(x)
    """
    return x.dtype == "O"



def _has_emails(x):
    if _is_string(x):
        check = is_email(x)
        return _np.any(check)
    else:
        return False


# Fiscal code
is_fiscal_code = _np.vectorize(lambda x: bool(_fc_re.match(x)))


def _has_fiscal_codes(x):
    if _is_string(x):
        check = is_fiscal_code(x)
        return _np.any(check)
    else:
        return False


# Telephone number:
is_telephone_number = _np.vectorize(lambda x: bool(_tel_re.match(x)))


def _has_telephone_numbers(x):
    if _is_string(x):
        check = is_telephone_number(x)
        return _np.any(check)
    else:
        return False


# Mobile number
is_mobile_number = _np.vectorize(lambda x: bool(_mobile_re.match(x)))


def _has_mobile_numbers(x):
    if _is_string(x):
        check = is_mobile_number(x)
        return _np.any(check)
    else:
        return False


def pii_find(x: _pd.DataFrame):
    """Find columns with probable piis (Personally Identifiable
    Informations) and return the colnames for further processing.

    >>> df = pd.DataFrame({
    >>>     "id" : [1,2,3],
    >>>     "cognome": ["brazorv", "gigetti", "ginetti"],
    >>>     "nome  " : [1,2,3],
    >>>     "mail": ["lgasd@asdkj.com", " asòdlk@asd.com", "aaaa"],
    >>>     "fc": ["nrgasd12h05h987z", "aaaa", "eee"],
    >>>     "num": ["0654-6540123", "aa", "eee"],
    >>>     "cel": ["3921231231", "aa", "eee"]
    >>>     })
    >>>
    >>> probable_piis = pii_find(df)
    >>> if probable_piis:
    >>>    df.drop(columns=probable_piis)

    """

    if not isinstance(x, _pd.DataFrame):
        msg = "x must be a pd.DataFrame"
        raise ValueError(msg)

    col = list(x.columns.values)

    # name and surname (looking at columns names)
    col_clean = [c.lower().strip() for c in col]
    surname_match = _columns_match(col_clean, ["cognome", "surname"])
    name_match = _columns_match(col_clean, ["nome", "name"])

    # finding other pii loking at data
    has_mails = [_has_emails(x[c]) for c in col]
    has_fcs = [_has_fiscal_codes(x[c]) for c in col]
    has_telephones = [_has_telephone_numbers(x[c]) for c in col]
    has_mobiles = [_has_mobile_numbers(x[c]) for c in col]

    probable_pii = []
    zipped = zip(
        col,
        surname_match,
        name_match,
        has_mails,
        has_fcs,
        has_telephones,
        has_mobiles,
    )
    for (
        var,
        is_surname,
        is_name,
        has_mail,
        has_fc,
        has_telephone,
        has_mobile,
    ) in zipped:
        pii_sum = (
            is_surname
            + is_name
            + has_mail
            + has_fc
            + has_telephone
            + has_mobile
        )
        if pii_sum:
            probable_pii.append(var)
            if is_surname:
                print(f"{var} matches 'surname'/'cognome'.")
            if is_name:
                print(f"{var} matches 'name'/'nome'.")
            if has_mail:
                print(f"{var} probably contains emails.")
            if has_fc:
                print(f"{var} probably contains fiscal codes.")
            if has_telephone:
                print(f"{var} probably contains telephone numbers.")
            if has_mobile:
                print(f"{var} probably contains mobile phones numbers.")
    return probable_pii


# -------------------------------------------------------------------------
# preprocessing varnames (from shitty excel): fix_columns
# -------------------------------------------------------------------------


def _compose(f, g):
    return lambda x: f(g(x))


def _replace_unwanted_chars(s):
    keep = _string.ascii_lowercase + _string.digits
    for ch in s:
        if ch not in keep:
            s = s.replace(ch, "_")
    return s


def _remove_duplicated_underscore(s):
    return _re.sub("_+", "_", s)


def _remove_external_underscore(s):
    return _re.sub("^_", "", _re.sub("_$", "", s))


def _add_x_if_first_is_digit(s):
    if s.startswith(tuple(_string.digits)):
        return "x" + s
    else:
        return s


def fix_varnames(x: str | list):
    """The good-old R preprocess_varnames, returns just the fixed strings. See
    sanitize_varnames for DataFrame or dict of DataFrames

    >>> fix_varnames("  asd 98n2 3")
    >>> fix_varnames([" 98n2 3", " L< KIAFJ8 0_________"])
    >>> fix_varnames(["asd", "foo0", "asd"])
    """
    if isinstance(x, str):
        original = [x]
    elif isinstance(x, list):
        original = x
    else:
        msg = "Only str and list are allowed, see sanitize_varnames for DataFrame or dict of DataFrames."
        raise ValueError(msg)
    # let's go functional: s is (should be) a str, the following are to be applied
    # in order
    funcs = [
        lambda s: str(s),
        lambda s: s.lower().strip(),
        lambda s: _replace_unwanted_chars(s),
        lambda s: _remove_duplicated_underscore(s),
        lambda s: _remove_external_underscore(s),
        lambda s: _add_x_if_first_is_digit(s),
    ]
    # sotto serve dato che compose applica prima la funzione a destra
    # e poi quella a sx (come in math) mentre vogliamo tenere l'ordine
    # di sopra
    funcs.reverse()
    worker = _functools.reduce(_compose, funcs)
    mod = [worker(s) for s in original]
    # handle duplicated names by adding numeric postfix
    has_duplicates = len(mod) != len(set(mod))
    if has_duplicates:
        seen = {}
        uniq = []
        for v in mod:
            if v not in seen:
                seen[v] = 0
                uniq.append(v)
            else:
                seen[v] += 1
                uniq.append(f"{v}_{seen[v]}")
    else:
        uniq = mod
    # se ci sono doppi nei nomi di partenza meglio evitare i dict se no i
    # doppi vengono considerati solo una volta
    # rename_dict = {}
    # for o, m in zip(original, uniq_mod):
    #     rename_dict.update({o: m})
    # if it was a string that was passed, return a string, not a list
    return uniq if not isinstance(x, str) else uniq[0]


def sanitize_varnames(x, return_tfd = True):
    """Fix a DataFrame or a dict of DataFrames names using fix_varnames to obtain
    the cleaned ones."""
    if isinstance(x, _pd.DataFrame):
        from_name = list(x.columns.values)
        to_name = fix_varnames(from_name)
        df = x.copy()
        df.columns = to_name
        tf = {t : f for t, f in zip(to_name, from_name)}
        if return_tfd:
            return df, tf
        else:
            return df
    elif isinstance(x, dict):
        dfs = {}
        tfs = {}
        for k, v in x.items():
            from_name = list(v.columns.values)
            to_name = fix_varnames(from_name)
            df = v.copy()
            df.columns = to_name
            tf = {t : f for t, f in zip(to_name, from_name)}
            dfs[k] = df
            tfs[k] = tf
        if return_tfd:
            return dfs, tfs
        else:
            return dfs



# -------------------------------------------------------------------------
# Coercion stuff below
# -------------------------------------------------------------------------


# decorators
def _verboser(f):
    def transformer(x: _pd.Series):
        coerced = f(x)
        report = _pd.DataFrame({"original": x, "coerced": coerced})
        new_na = (~_pd.isna(x)) & (_pd.isna(coerced))
        if _np.any(new_na):
            print("Please check if the following coercions are ok:")
            print(report[new_na])
        return coerced
    return transformer


# --------------- coercion workers ----------------------------------------

def to_bool(x: _pd.Series):
    """Coerce to boolean a pd.Series (numeric) using astype keeping NAs as NAs

    >>> to_bool(pd.Series([1,0,1,0]))
    >>> to_bool(pd.Series([1,0.,1,0.]))
    >>> to_bool(pd.Series([1,0,1,0, np.nan]))
    """
    nas = _pd.isna(x)
    rval = x.astype("boolean")
    rval[nas] = _pd.NA
    return rval


def _replace_comma(x: _pd.Series):
    if _is_string(x):
        # nas = x.isin(["", _pd.NA, _np.nan])
        nas = (x.isna()) | (x == "")
        rval = x.astype("str").str.replace(",", ".")
        rval[nas] = _pd.NA
        return rval
    else:
        return x


def to_integer(x: _pd.Series):
    """Coerce a pd.Series to integer (if possible)

    >>> to_integer(pd.Series([1., 2., 3., 4., 5., 6., np.nan]))
    >>> to_integer(pd.Series(["2001", "2011", "1999", ""]))
    >>> to_integer(pd.Series(["1.1", "1,99999", "foobar"])) # fails because of 1.99
    """
    s = _replace_comma(x)
    # return _np.floor(_pd.to_numeric(s, errors='coerce')).astype('Int64')
    # mi fido piu di quella di sotto anche se fallisce con i numeri con virgola
    # (che ci può stare, per i quale bisogna usare np.floor)
    return _pd.to_numeric(s, errors="coerce").astype("Int64")


def to_numeric(x: _pd.Series):
    """Coerce a pd.Series using pd.to_numeric

    >>> to_numeric(pd.Series([1, 2, 3]))
    >>> to_numeric(pd.Series([1., 2., 3., 4., 5., 6.]))
    >>> to_numeric(pd.Series(["2001", "2011", "1999"]))
    >>> to_numeric(pd.Series(["1.1", "2,1", "asd", ""]))
    """
    s = _replace_comma(x)
    return _pd.to_numeric(s, errors="coerce").astype("Float64")


def to_datetime(x: _pd.Series):
    """Coerce to a datetime a pd.Series

    >>> import datetime as dt
    >>> to_datetime(pd.Series([str(dt.datetime.now())] * 6))
    """
    return _pd.to_datetime(x, errors="coerce")


def to_date(x: _pd.Series):
    """Coerce to a date a pd.Series

    >>> import datetime as dt
    >>> to_date(pd.Series([str(dt.datetime.now())] * 6))
    >>> to_date(pd.Series(["2020-01-02", "2021-01-01", "2022-01-02"] * 2))
    """
    return to_datetime(x).dt.floor("D")


_dates_re = _re.compile(r"[^/\d-]") # keep only numbers, - and /, and just hope for the best

def _extract_dates_worker(x):
    if isinstance(x, str): # handle missing values (sono float)
        polished = _dates_re.sub("", x)
        return _pd.to_datetime(polished)
    else:
        # return _np.nan
        return _pd.NA


def extract_dates(x):
    """Try to extract dates from shitty strings and convert them to proper

    >>> extract_dates(pd.Series(["2020-01-02", "01/01/1956", "asdasd 12-01-02"] * 2))
    """
    return x.apply(_extract_dates_worker)


def to_categorical(x: _pd.Series,
                   categories: list[str] = None,
                   lowcase: bool = False):
    """Coerce to categorical a pd.Series, with blank values as missing

    >>> to_categorical(pd.Series([1, 2, 1, 2, 3]))
    >>> to_categorical(pd.Series([1, 2., 1., 2, 3, np.nan]))
    >>> to_categorical(pd.Series(["AA", "sd", "asd", "aa", "", np.nan]))
    >>> to_categorical(pd.Series(["AA", "sd", "asd", "aa", "", np.nan]), categories=["aa", "AA"]  )
    >>> to_categorical(pd.Series(["AA", "sd", "asd", "aa", ""]), lowcase = True)
    """
    if _is_string(x):
        # string preprocessing
        # rm spaces and uniform NAs
        x = x.str.strip()
        nas = (x.isna()) | (x == "")
        x[nas] = _pd.NA
        if lowcase:
            x = x.str.lower()
            if categories != None:
                categories = [c.lower() for c in categories]

    # categorical making
    return _pd.Categorical(x, categories=categories)


def to_noyes(x: _pd.Series):
    """Coerce to no/yes a string pd.Series

    >>> to_noyes(pd.Series(["","yes","no","boh", "si", np.nan]))
    >>> to_noyes(pd.Series([1,0,0, np.nan]))
    >>> to_noyes(pd.Series([1.,0.,0., np.nan]))
    """
    if _is_string(x):
        # take only the first character and map to n/y
        tmp = x.str.strip().str.lower().str[0]
        tmp[tmp == "s"] = "y"
        tmp = tmp.replace({"0": "n", "1": "y"}) # 0/1 for strings
    else:
        # try to convert to boolean and map to n/y
        tmp = to_bool(x).map({False: "n", True: "y"})


    return to_categorical(tmp.map({"n": "no", "y": "yes"}),
                          categories=["no", "yes"])




def to_sex(x: _pd.Series):
    """Coerce to male/female a pd.Series of strings (Mm/Ff)

    >>> to_sex(pd.Series(["","m","f"," m", "Fm", np.nan]))
    """
    if not _is_string(x):
        msg = "to_sex only for strings vectors"
        raise Exception(msg)

    # take the first letter (Mm/Ff)
    tmp = x.str.strip().str.lower().str[0]
    tmp = tmp.map({"m": "male", "f": "female"})
    return to_categorical(tmp, categories=["male", "female"])



def to_recist(x: _pd.Series):
    """Coerce to recist categories a pd.Series of strings

    >>> to_recist(pd.Series(["RC", "PD", "SD", "PR", "RP", "boh", np.nan]))
    """
    if not _is_string(x):
        msg = "to_recist only for strings vectors"
        raise Exception(msg)

    # rm spaces and uppercase and take the first two letters
    tmp = x.str.strip().str.upper().str[:2]
    # uniform italian to english
    ita2eng = {"RC": "CR", "RP": "PR"}
    return to_categorical(tmp.replace(ita2eng),
                          categories=["CR", "PR", "SD", "PD"])


def to_other_specify(x: _pd.Series):
    """Try to polish a bit a free-text variable and create a categorical one

    >>> x = pd.Series(["asd", "asd", "", "prova", "ciao", 3]+ ["bar"]*4)
    >>> to_other_specify(x)
    """
    nas = (x.isna()) | (x == "")
    tmp = x.copy().astype("str").str.strip().str.lower()
    tmp[nas] = _pd.NA
    # categs ordered by decreasing counts
    categs = list(tmp.value_counts().index)
    return to_categorical(tmp, categories=categs)


def to_string(x: _pd.Series):
    """Coerce the pd.Series passed to a string Series

    >>> x = pd.Series(["asd", "asd", "", "prova", "ciao", 3]+ ["bar"]*4)
    >>> to_string(x)
    """
    nas = x.isna()
    rval = x.astype("str")
    rval[nas] = _pd.NA
    return rval


# --------------- coercion class ----------------------------------------
class Coercer:
    """Coercer of a pandas DataFrame.

    Given directives as a dict (variable name as key, function/coercer
    as value) it applies all the function on a copy of the DataFrame
    and return it.  If verbose print a report of the introduced
    missing values with the coercion for check

    >>> from pylbmisc.dm import *
    >>> import pandas as pd
    >>> import numpy as np
    >>>
    >>> raw = pd.DataFrame({
    >>>     "idx" :  [1., 2., 3., 4., 5., 6., "2,0", "", np.nan],
    >>>     "sex" :  ["m", "maschio", "f", "female", "m", "M", "", "a", np.nan],
    >>>     "now":   [str(dt.datetime.now())] * 6 + [np.nan, "", "a"],
    >>>     "date":  ["2020-01-02", "2021-01-01", "2022-01-02"] * 2  + [np.nan, "", "a"],
    >>>     "state": ["Ohio", "Ohio", "Ohio", "Nevada", "Nevada", "Nevada"] + [np.nan, "", "a"],
    >>>     "ohio" : [1, 1, 1, 0, 0, 0] + [np.nan] * 3 ,
    >>>     "year":  [str(y) for y in [2000, 2001, 2002, 2001, 2002, 2003]] + [np.nan, "", "a"],
    >>>     "pop":   [str(p) for p in [1.5, 1.7, 3.6, np.nan, 2.9]]  + ["3,2", np.nan, "", "a"],
    >>>     "recist" : ["", "pd", "sd", "pr", "rc", "cr"] + [np.nan, "", "a"],
    >>>     "other" : ["b"]*3 + ["a"]*2 + ["c"] + [np.nan, "", "a"]
    >>> })
    >>> 
    >>> directives_new = {
    >>>     lb.dm.to_integer: ["idx", "year"],
    >>>     lb.dm.to_sex : ["sex"],
    >>>     lb.dm.to_datetime : ["now"],
    >>>     lb.dm.to_date: ["date"],
    >>>     lb.dm.to_categorical: ["state"],
    >>>     lb.dm.to_bool: ["ohio"],
    >>>     lb.dm.to_numeric: ["pop"],
    >>>     lb.dm.to_recist: ["recist"],
    >>>     lb.dm.to_other_specify: ["other"]
    >>> }
    >>> 
    >>> cleaned1 = lb.dm.Coercer(raw, fv = directives_new).coerce()
    >>> 
    >>> raw
    >>> cleaned1
    >>> 
    >>> # ------------------------------------------------
    >>> # OLD API
    >>> # ------------------------------------------------
    >>>
    >>> directives_new2 = {
    >>>     "lb.dm.to_categorical": ["state"],
    >>>     "lb.dm.to_date": ["date"],
    >>>     "lb.dm.to_datetime" : ["now"],
    >>>     "lb.dm.to_integer": ["idx", "year"],
    >>>     "lb.dm.to_noyes": ["ohio"],
    >>>     "lb.dm.to_numeric": ["pop"],
    >>>     "lb.dm.to_other_specify": ["other"],
    >>>     "lb.dm.to_recist": ["recist"],
    >>>     "lb.dm.to_sex" : ["sex"]
    >>> }
    >>>
    >>>
    >>> coercer2 = lb.dm.Coercer(raw, fv = directives_new2)
    >>> cleaned2 = coercer2.coerce()
    >>>
    """

    def __init__(
        self,
        df: _pd.DataFrame,
        fv: dict | None = None, # function: variables dict
        verbose: bool = True,
    ):
        self._df = df
        self._verbose = verbose
        if fv == None:
            msg = "fv can't be None"
            raise ValueError(msg)
        else:
            # Experimental below
            parent_frame = _inspect.currentframe().f_back
            # put it in the vf_dict format evaluating the f in
            # parent frame variable dict
            reversed = {}
            for f, vars in fv.items():
                # if f is a string change it to function taking from the enclosing
                # environment
                f = (
                    eval(f, parent_frame.f_locals, parent_frame.f_globals)
                    if isinstance(f, str)
                    else f
                )
                for v in vars:
                    reversed.update({v: f})
            self._directives = reversed

    def coerce(self, keep_only_coerced=False) -> _pd.DataFrame:
        # do not modify the input data
        df = self._df.copy()
        directives = self._directives
        verbose = self._verbose
        # make verbose all the functions by decorating them
        if verbose:
            for var in directives.keys():
                directives[var] = _verboser(directives[var])
        # apply the coercers, but first modify the pd printing options temporarily to
        # handle long reporting of changes
        old_nrows = _pd.get_option("display.max_rows")
        _pd.set_option("display.max_rows", None)
        for var, fun in directives.items():
            if var not in df.columns:
                msg = f"{var} not in df.columns, aborting."
                raise ValueError(msg)
            if verbose:
                print(f"Processing {var}.")
            df[var] = fun(df[var])
        _pd.set_option("display.max_rows", old_nrows)
        # return results
        if keep_only_coerced:
            vars = list(directives.keys())
            df = df[vars]
        return df



import argparse
import asyncio
import telegram  # pip install --user python-telegram-bot
from pathlib import Path
from ..tg import bot_token, user_id, group_id


async def main(what, to):
    # handling text or attachment
    what_path = Path(what).resolve()
    send = "file" if what_path.exists() else "msg"
    if send == "file":
        ext = what_path.suffix.lower()
    else:
        msg = what

    winston_token = bot_token("winston_lb_bot")
    winston = telegram.Bot(winston_token)
    async with winston:
        if send == "msg":
            await winston.send_message(chat_id=to, text=msg)
        elif ext in {".mp3", ".m4a"}:
            await winston.send_audio(chat_id=to, audio=what_path)
        elif ext in {".png", ".jpg", ".jpeg"}:
            await winston.send_photo(chat_id=to, photo=what_path)
        elif ext in {".mp4"}:
            await winston.send_video(chat_id=to, video=what_path)
        else:
            await winston.send_document(chat_id=to, document=what_path)


def winston_sends():
    """
    winston_sends "ciao" user::lucailgarb
    winston_sends file.pdf user::lucailgarb
    winston_sends file.png group::da_salvare
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("what")
    parser.add_argument("to")
    args = parser.parse_args()
    what = args.what
    to = args.to
    to_spl = to.split("::")
    to_type = to_spl[0]
    to_label = to_spl[1]
    if to_type == 'user':
        to_id = user_id(to_label)
    elif to_type == 'group':
        to_id = group_id(to_label)
    else:
        msg = "Currently only user::* and group::* parsed."
        raise ValueError(msg)
    asyncio.run(main(what, to_id))



"""Statistical utilities/routines"""

import pandas as _pd
import numpy as _np
from .utils import match_arg as _match_arg


def _pstar_worker(p):
    if p < 0.001:
        return "***"
    elif p < 0.01:
        return "**"
    elif p < 0.05:
        return "*"
    else:
        return ""


def _pformat_worker(p):
    if p < 0.001:
        return "< 0.001"
    elif p == 1:
        return "0.999"
    else:
        return "%.3f" % p


def p_star(p):
    """The unholy p-value stars"""
    if isinstance(p, float):
        return _pstar_worker(p)
    elif isinstance(p, _pd.Series):
        return p.map(_pstar_worker)
    else:
        return None


def p_format(p):
    """Pretty print (format) p-value for publication"""
    if isinstance(p, float):
        return _pformat_worker(p)
    elif isinstance(p, _pd.Series):
        return p.map(_pformat_worker)
    else:
        return None


def p_adjust(p, method="holm"):
    """A quick/simple port of p.adjust with main methods used by me

    See scipy.stats for FDR/BH methods

    # without missing
    # > wikipedia = c(0.01, 0.04, 0.03, 0.005)
    # > p.adjust(wikipedia)
    # [1] 0.03 0.06 0.06 0.02
    >>> wikipedia = [0.01, 0.04, 0.03, 0.005]
    >>> p_adjust(wikipedia)
    # with missing
    # > wikipedia_miss = c(0.01, NA, 0.04, 0.03, 0.005, NA)
    # > p.adjust(wikipedia_miss)
    # [1] 0.03   NA 0.06 0.06 0.02   NA
    >>> wikipedia_miss = [0.01, np.nan, 0.04, 0.03, 0.005, np.nan]
    >>> p_adjust(wikipedia_miss)
    """
    if isinstance(p, list):
        x = _np.array(p)
    elif isinstance(p, _pd.Series):
        x = p.values
    elif isinstance(p, _np.ndarray):
        x = p
    else:
        msg = "x must be list, series or array"
        raise ValueError(msg)

    # checking content type
    if not _np.issubdtype(x.dtype, _np.floating):
        msg = "p-values must be a float."
        raise ValueError(msg)

    # checking method requested
    allowed_methods = ["none", "bonferroni", "holm"]
    method = _match_arg(method, allowed_methods)

    # methods implementation
    n = len(x)
    n_nonmiss = (~_np.isnan(x)).sum()
    if method == "none":
        p_adj = x
    elif method == "bonferroni":
        p_adj = x * n_nonmiss
        for i, p in enumerate(p_adj):
            if p > 1:
                p_adj[i] = 1
    elif method == "holm":
        p_adj = _np.empty(n)
        values = [(pval, prog_id) for prog_id, pval in enumerate(x)]
        # ordino sulla base dei p.value, i missing finiscono alla fine ma ho
        # tenuto l'ordine dove compaiono nel secondo elemento
        # ora per poter ordinare sulla base dei p, in presenza di missing è necessario
        # usare np.sort che richiede uno structured array (vedi np.sort)
        dtype = [('pval', float), ('prog_id', int)]
        values = _np.array(values, dtype=dtype)
        sa = _np.sort(values, order='pval')
        pmax = 0  # initialize so the first maximum
        for k, vals in enumerate(sa, start=1):
            pval, prog_id = vals
            if not _np.isnan(pval):
                res = pval * (n_nonmiss - k + 1)
                res = res if res < 1 else 1
                pmax = res if res > pmax else pmax
                p_adj[prog_id] = pmax
            else:
                p_adj[prog_id] = _np.nan
    else:
        msg = "method uncorrectly specified"
        raise ValueError(msg)

    # exiting
    return p_adj



# SPDX-FileCopyrightText: 2023-present Luca Braglia <lbraglia@gmail.com>
#
# SPDX-License-Identifier: MIT

from pylbmisc import dm, io, iter, stats, tg, tmpl, utils

__all__ = ["dm", "io", "iter", "stats", "tg", "tmpl", "utils"]



import unittest
import pandas as pd
import numpy as np
from pylbmisc.dm import to_bool, to_integer, to_numeric, to_datetime, to_date, to_categorical, to_noyes, to_sex, to_recist, to_other_specify, to_string

class TestDMFunctions(unittest.TestCase):

    def test_to_bool(self):
        series = pd.Series([1., 0, 1., 0, np.nan])
        expected = pd.Series([True, False, True, False, pd.NA], dtype="boolean")
        result = to_bool(series)
        pd.testing.assert_series_equal(result, expected)

    def test_to_integer_float(self):
        series = pd.Series([1.0, 2.0, 3.0, 4.0, np.nan])
        expected = pd.Series([1, 2, 3, 4, pd.NA], dtype="Int64")
        result = to_integer(series)
        pd.testing.assert_series_equal(result, expected)

    def test_to_integer_string(self):
        series = pd.Series(["2001", "2011", "1999", np.nan])
        expected = pd.Series([2001, 2011, 1999, pd.NA], dtype="Int64")
        result = to_integer(series)
        pd.testing.assert_series_equal(result, expected)

    def test_to_numeric(self):
        series = pd.Series(["1.1", "2,1", "3.0", "4.5", ""])
        expected = pd.Series([1.1, 2.1, 3.0, 4.5, np.nan], dtype="Float64")
        result = to_numeric(series)
        pd.testing.assert_series_equal(result, expected)

    def test_to_datetime(self):
        series = pd.Series(["2020-01-01", "2021-01-01", "2022-01-01", "", np.nan])
        expected = pd.to_datetime(pd.Series(["2020-01-01", "2021-01-01", "2022-01-01", pd.NaT, pd.NaT]))
        result = to_datetime(series)
        pd.testing.assert_series_equal(result, expected)

    def test_to_date(self):
        series = pd.Series(["2020-01-01 12:34:56", "2021-01-01 00:00:00", "2022-01-01 23:59:59", "", np.nan])
        expected = pd.to_datetime(pd.Series(["2020-01-01", "2021-01-01", "2022-01-01", pd.NaT, pd.NaT])).dt.floor("D")
        result = to_date(series)
        pd.testing.assert_series_equal(result, expected)

    def test_to_categorical(self):
        series = pd.Series(["A", "B", "A", "C", "", np.nan])
        expected = pd.Categorical(["A", "B", "A", "C", pd.NA, pd.NA])
        result = to_categorical(series)
        # categorical in pandas seems to be extension arrays
        # https://pandas.pydata.org/community/blog/extension-arrays.html
        pd.testing.assert_extension_array_equal(result, expected)

    def test_to_noyes(self):
        series = pd.Series(["","yes","no","boh", "si", np.nan])
        expected = pd.Categorical([pd.NA, "yes", "no", pd.NA, "yes", pd.NA], categories=["no", "yes"])
        result = to_noyes(series)
        pd.testing.assert_extension_array_equal(result, expected)

    def test_to_sex(self):
        series = pd.Series([""    ,"m","f"," m", "Fm", np.nan])
        expected = pd.Categorical([pd.NA , "male", "female", "male", "female", pd.NA],
                                  categories=["male", "female"])
        result = to_sex(series)
        pd.testing.assert_extension_array_equal(result, expected)

    def test_to_recist(self):
        series = pd.Series(["RC", "PD", "SD", "PR", "RP", "boh", np.nan])
        expected = pd.Categorical(["CR", "PD", "SD", "PR", "PR", pd.NA, pd.NA],
                                  categories=["CR", "PR", "SD", "PD"])
        result = to_recist(series)
        pd.testing.assert_extension_array_equal(result, expected)

    def test_to_other_specify(self):
        series = pd.Series(["foo", "bar", "foo", "baz", ""])
        expected = pd.Categorical(["foo", "bar", "foo", "baz", pd.NA],
                                  categories=["foo", "bar", "baz"])
        result = to_other_specify(series)
        pd.testing.assert_extension_array_equal(result, expected)

    def test_to_string(self):
        series = pd.Series([1, 2, 3, 4, np.nan]).astype("Int32")
        expected = pd.Series(["1", "2", "3", "4", pd.NA])
        result = to_string(series)
        pd.testing.assert_series_equal(result, expected)

if __name__ == '__main__':
    unittest.main()



"""Telegram related utilities"""

import os as _os
import json as _json
import pandas as _pd

from pathlib import Path as _Path


def chat2df(fpath: str | _Path) -> _pd.DataFrame:
    """Telegram chat json to pandas DataFrame

    Trasforma una chat json esportata dal clienti desktop in un
    DataFrame Pandas.

    Args:
       fpath (str|Path): path ad un file json esportato da Telegram

    Returns:
       pandas.DataFrame: the chat

    """
    fpath = _Path(fpath)
    with fpath.open() as f:
        js = _json.load(f)
        msg = js["messages"]
        return _pd.DataFrame(msg)


def bot_token(b):
    """
    Get Bot API token from environment variables.
    """
    return _os.environ["TG_BOT_%s" % b]


def user_id(u):
    """
    Get user id from environment variables.
    """
    return _os.environ["TG_USER_%s" % u]


def group_id(g):
    """
    Get group id from environment variables.
    """
    return _os.environ["TG_GROUP_%s" % g]



ctsu = r"""

\begin{table}
  \centering
  \begin{tabular}{ll}
    \hline
    \textbf{Titolo dello studio} & \\
    \textbf{Acronimo} & \\
    \textbf{Versione protocollo} & 1.0 \\
    \textbf{Data rilascio protocollo} &  \\
    \textbf{Codici registrazione} & se applicabili (es ClinicalTrials.gov: NCT01066572) \\
    \hline
    \textbf{Promotore} & Azienda farmaceutica, gruppo cooperativo, ente non profit che promuove lo studio. Nel caso di studi spontanei locali inserire Azienda USL - IRCCS di Reggio Emilia.\\
    \textbf{Proponente} &  Nome e qualifica delle persone autorizzate a firmare il protocollo e gli emendamenti a nome del promotore. Data e firma, per accettazione.\\
    \textbf{Partecipanti} &  Sperimentatori responsabili della conduzione dello studio con relativa strutture di appartenenza  ( es.: strutture specialistiche e laboratori ...).\\
    \textbf{Recapiti randomizzazione} &  \\
    \textbf{Referenti farmacovigilanza} &  \\
    \textbf{Referente statistico} &  \\
    \textbf{Referente monitoraggio} &  \\
  \end{tabular}
\end{table}


 \\
 \\
 \\

\textbf{UNA VOLTA COMPLETATA LA SCRITTURA INSERIRE QUI IL SOMMARIO
  (eventualmente in pagina a sé stante)}

 \\
 \\
 \\


% ---------------------------------------------------------
\section{Background}
Fornire una sintesi sulla problematica affrontata (es la patologia),
sulla attuale pratica clinica (es da linee guida), e sulle questioni
attualmente aperte (con riferimento aggiornato alla letteratura
esistente).

% ---------------------------------------------------------
\section{Lo studio}

\subsection{Ipotesi e razionale}
L'intento dello studio deve essere definito in modo chiaro e
preciso. Specificare qui la/le ipotesi da verificare e le ragioni che
giustificano tale proposta (es non precedentemente/sufficientemente
indagate in letteratura?).

\subsection{Disegno dello studio}
Sintetizzare gli aspetti salienti dello studio. Ad esempio la
tipologia (sperimentale/osservazionale); mono/multicentricità;
retrospettivo/prospettico; lo tipologia/scopo dell'indagine
(descrittivo, diagnostico, di trattamento, prognostico ecc).\\
Nel caso di studi su un trattamento, il metodo di allocazione dello
stesso (randomizzata/non randomizzata); la tipologia di trattamento
(farmacologico, dispositivo, altro); la fase (I-II-III-IV); gli
aspetti di mascheramento (in aperto, cieco, doppio cieco, triplo
cieco); ecc.\\
Infine, qualora presenti, quali studi ancillari e/o sotto-studi sono
previsti.

\section{Popolazione}
Descrivere qui la tipologia di contesto (es ospedaliero/territoriale)
dove si realizzerà lo studio e dove avverrà l'arruolamento (o
comunque, anche nel caso degli studi retrospettivi, il contesto a cui
si riferiranno i i risultati dello studio).

\subsection{Criteri di inclusione}
Inserirli in forma schematica (elenchi puntati/numerati) e
dettagliati.

\subsection{Criteri di esclusione}
Sempre in forma schematica (elenchi puntati/numerati) e dettagliati;
non serve introdurre negazioni dei criteri di inclusione (ad esempio
se un criterio di inclusione è età >= 18 qui non serve specificare età
< 18).


\subsection{Modalità di reclutamento dei pazienti}
INSERIRE QUALCHE SUGGERIMENTO GENERICO / SEMPLICE QUI su cosa inserire
qui?

\subsection{Allocazione al trattamento (randomizzazione)}
Nel caso di studi su trattamento, descrivere il processo di
assegnazione di ciascun paziente al trattamento (es randomizzazione o
altro).


\section{Aspetti organizzativi e logistici}


\subsection{Fasi e durata dello studio}
Sintetizzare qui, anche mediante esemplificazioni grafiche (es
flow-chart, tempogrammi, diagrammi GANTT) la pianificazione dello
studio del suo complesso e il susseguirsi delle varie sue fasi: ad
esempio arruolamento, trattamento, follow up (FPFV-LPLV: da first
patient first visit a last patient last visit) e raccolta dati, analisi).\\
A valle di questo, fornire una stima della durata complessiva
prevista.


\subsection{Programma di trattamento}
Dosaggi e schedule, durata, tossicità attese, modifiche al
trattamento.

\subsection{Valutazione outcomes ed effetti collaterali}
Tempistiche e criteri/standard adottati (es RECIST) per la valutazione
degli outcome e degli effetti avversi (se applicabile).

\section{Interruzione dello studio}
Inserire le norme di interruzione per lo studio e criteri per il
ritiro dei pazienti dallo stesso.

% ---------------------------------------------------------
\section{Obiettivi}
Riportare in questa sezione le ipotesi da sottoporre a verifica,
scritte in maniera schematica e gerarchica (un obiettivo primario e
alcuni obiettivi secondari).


\paragraph{Primario}
Ad esempio: Obiettivo primario è valutare l'effetto del trattamento
sulla sopravvivenza dei pazienti randomizzati.

\paragraph{Secondari}


\section{Aspetti statistici}

\subsection{Outcome}
Riportare in questa sezione le variabili che verranno utilizzate per
valutare le ipotesi. Come sono definite e quando vengono misurate.\\
Ad esempio: per la valutazione della sopravvivenza si utilizzerà la
Progression Free Survival, definita come tempo tra la randomizzazione
e la prima tra data di ripresa o decesso.

\paragraph{Primario}


\paragraph{Secondari}

\subsection{Analisi}
Questa sezione è a cura dell'Ufficio Studi Clinici e Statistica.

\paragraph{Primaria}

\paragraph{Secondarie}

\subsection{Ampiezza campionaria}
Questa sezione è a cura dell'Ufficio Studi Clinici e Statistica.

\section{Aspetti di data management}
Specificare quali strumenti saranno impiegati per la raccolta dei dati
(es Excel/Smarty). Qualora i dati derivino dall'estrazione a partire
da una/più fonte informativa/e già esistente/i (es cartella clinica
informatizzata, registri/database, etc) specificare la/e
fonte/i. Infine fare riferimento ad un documento allegato, per quanto
riguarda l'elenco delle variabili raccolte o la struttura della CRF.


\section{Aspetti etici e autorizzativi}
Specificare aspetti quali:
\begin{itemize}
\item considerazioni etiche e riferimenti normativi;
\item informazione sullo studio, modalità di raccolta del
  consenso allo studio ed al trattamento dati;
\item aspetti assicurativi;
\item accordi finanziari (trattati in documento separato);
\item accesso ai dati da parte di autorità regolatorie.
\end{itemize}

\section{Monitoraggio e QC}
Specificare aspetti quali:
\begin{itemize}
\item piano di monitoraggio proposto;
\item modalità di accesso ai dati e controlli di qualità.
\end{itemize}

\section{Sicurezza/Farmacovigilanza}
Modalità di rilevazione e segnalazione degli eventi, obblighi di
farmacovigilanza, SOP di riferimento.

\section{Responsabilità e politica di pubblicazione}
Criteri di pubblicazione (se non trattati in documento
separato). Proprietà dei dati e modalità di diffusione.

\section{Riferimenti bibliografici}
Inserire l'elenco della bibliografia citata nel documento.

\section{Appendici e allegati}
Eventuali documenti in allegato (es.: strumenti di valutazione,
questionari, scale, schede segnalazione eventi avversi seri, schede
raccolta dati, ecc)

"""

heap = r"""

\section{Administrative information}
This Health Economics Analysis Plan (HEAP) is written according to
\cite{thorn2021heap}.


\begin{table}
  \centering
  \begin{tabular}{ll}
    \hline
    \textbf{Title} &
                     Title that matches protocol and which includes the phrase ‘Health Economics
                     Analysis Plan'
                     % Health economics analysis plan for the [trial name]: a
                     % multi-centre randomised controlled trial to assess whether X
                     % improves the outcome of Y
    \\
    \textbf{Trial registration number} &
                                         Trial registration number and name of registry that
                                         uniquely identifies the clinical trial on a publicly
                                         accessible registry (and other relevant trial study numbers)
                                         % ISRCTNXXXXXXXX (ISRCTN
                                         % registry)
    \\
    \textbf{Source of funding} &
                                 Name of funders for trial and economic evaluation and funder(s)’ reference number(s)
                                 % National Institute for Health
                                 % Research, Health Technology
                                 % Assessment; Reference number –
                                 % XX/XX/XX
    \\
    \textbf{Purpose of HEAP} &
                               Brief statement of the purpose of the HEAP
                               % The purpose of this HEAP is to
                               % describe the analysis and reporting
                               % procedure intended for the economic
                               % analyses to be undertaken. The
                               % analysis plan is designed to ensure
                               % that there is no conflict with the
                               % protocol and associated statistical
                               % analysis plan and it should be read
                               % in conjunction with them.

    \\
    \textbf{Trial protocol version} &
                                      Trial protocol version number associated with this HEAP
                                      % This document has been written
                                      % based on information contained
                                      % in the trial protocol version
                                      % 4, dated 14 October 2019
    \\
    \textbf{Trial Statistical Analysis Plan (SAP) version} &
                                                             SAP version number associated with this HEAP
                                                             % SAP Version: 1.0, Date: 4 August 2019
    \\
    \textbf{Trial HEAP version} &
                                  Sequential number and date of this version
                                  % HEAP Version: 1.0, Date: 4 August 2019
    \\                          
    \textbf{HEAP revisions} &
                              Date, justification for revision and summary of changes to the HEAP. Specify the individual making any revisions/changes to the HEAP.
                              % The HEAP revision history can be
                              % presented in tabular format with the
                              % following column headings: Updated
                              % HEAP Version No; Protocol version;
                              % Section number changed; Description
                              % of, and reason for, change; Individual
                              % making the change (e.g. Dr [name] made
                              % the revision before any comparative
                              % analyses had been conducted); Date
                              % changed.
                              % Each row subsequently added to the
                              % table will indicate each HEAP revision
                              % change.
    \\
    \textbf{Roles and responsibilities} &
                                          Names, affiliations and roles of individuals who have significantly contributed to the HEAP
                                          % This HEAP was prepared by
                                          % Dr [name] (junior health
                                          % economist) and approved
                                          % by, Prof/Dr [name] (senior
                                          % health economist). The
                                          % trial health economist(s)
                                          % [name (s)] are responsible
                                          % for conducting and
                                          % reporting the economic
                                          % evaluation in accordance
                                          % with the HEAP.
    \\
    \textbf{Signature(s) of person(s) writing HEAP} &
                                                      Signature(s) of the person(s) writing the HEAP (and date)
    \\
    \textbf{Signature of senior health economist} &
                                                    Signature of senior health economist who is guarantor of the economic evaluation (and date)
    \\
    \textbf{Signature of Chief Investigator} &
                                               Signature of the Chief Investigator for the trial (and date)
    \\
    \hline
  \end{tabular}
\end{table}

\subsection*{Table  of contents}

List of HEAP contents with page numbers

\subsection*{Abbreviations/glossary of terms/definitions}

\begin{table}
  \centering
  \begin{tabular}{ll}
    \hline
    CEA & Cost effectiveness analysis \\
    CUA & Cost utility analysis \\
    QALY & quality-adjusted life year \\
    NHS & National Health Service \\
    \hline
  \end{tabular}
\end{table}



\section{Trial introduction \& background}

\subsection{Trial background and rationale}

Synopsis of trial background and rationale including a brief
description of research question and brief justification for
undertaking the trial

% Testing the effectiveness of new school-based educational
% interventions for children and young people with depression is a
% priority. The [trial name] is a pragmatic, parallel design, two arm,
% multicentre randomised controlled trial to determine whether a
% group-based CBT course is more effective than curriculum as usual
% for people who have had a diagnosis of depression in the past 24
% months.

\subsection{Aim(s) of the trial}

Clearly and briefly state the main aim(s) of the trial

% Briefly, the [trial name] RCT aims to inform primary care
% prescribing practice by investigating the severity and duration of
% depressive symptoms that are associated with a clinically
% significant response to sertraline compared to placebo, in people
% presenting to primary care with depression.


\subsection{Objectives and/or research hypotheses of the trial}

Describe specific trial objectives (primary and secondary) or trial
hypotheses

% Primary objective: to assess whether there is a clinically important
% difference in the impact of fatigue 26 weeks after baseline
% (i.e. starting the intervention) between patients participating in a
% cognitive-behavioural course delivered by the clinical team compared
% with patients receiving usual care. Secondary objectives: to compare
% differences between groups for secondary outcomes of fatigue
% severity, sleep, pain, disability, health service use, quality of
% life and cost-effectiveness for the NHS, patients and society.


\subsection{Trial population}

Describe the trial inclusion and exclusion criteria

% Inclusion criteria: GP patients with a diagnosis of severe mental
% illness. Raised total cholesterol above 5 or raised total
% cholesterol above 4 and one or more of the following risk factors:
% Current smoker, BMI>30kg/m2, diagnosis of diabetes and aged 30-75
% years old.  Exclusion criteria: Under acute psychiatric care.  Life
% expectancy of less than 6 months.  Pregnant at baseline.  Unable to
% give informed consent.


\subsection{Intervention(s) and comparator(s)}

Describe the intervention(s) and comparator(s)

% Intervention: Patients in the GP practices allocated to the
% intervention will be offered an enhanced cardiovascular disease risk
% reducing service over 6 months. The intervention will include one or
% more of the following elements: lifestyle advice, education on diet
% and exercise, behavioural techniques to encourage patients to make
% healthy lifestyle changes, signposting to smoking cessation or
% physical activity programmes.  Comparator: Treatment as usual.


\subsection{Trial design}

Briefly describe the trial design including type of trial such as
cluster, crossover, etc. Can also include details of power
calculation, sample size (including any separate calculations for
economic endpoints), randomisation and blinding.

% This is a non-blinded cluster RCT with GP practices being randomised
% to the intervention or treatment as usual on a 1:1 basis.
% Randomisation of GP practices will be carried out by an independent
% statistician. A total of 500 patients from 75 GP practices will be
% recruited into the trial.


\subsection{Trial start and end dates}

Trial recruitment start and end dates and the follow-up period

% Recruitment started in October 2017 and is due to finish in October
% 2020. The follow-up period will run for 12 months until October
% 2021.

\section{Economic approach/overview}

\subsection{Aim(s) of economic evaluation}

Describe the aim(s) of the economic evaluation

% The aim of the economic evaluation is to address the question “What
% is the cost-effectiveness of continuous subcutaneous insulin
% injections (CSII) compared with multiple daily injections (MDI) in
% patients with type 1 diabetes receiving a structured education
% programme?”.

\subsection{Objective(s) of economic evaluation}

Describe the objectives (primary and secondary) of the economic evaluation

% The primary objective of the health economic evaluation is to
% estimate the long-term cost-effectiveness of continuous subcutaneous
% insulin injections (CSII) versus multiple daily injections (MDI) for
% patients receiving a structured education programme over a life time
% horizon using economic modelling techniques. A secondary objective
% is to calculate the short-term cost-effectiveness of CSII versus MDI
% in the first 2 years after patients receive the education programme
% in a within-trial economic evaluation.

\subsection{Overview of economic analysis}

Briefly outline and justify the type of economic evaluation to be
undertaken, identifying the primary economic analysis and outlining
the analysis plan and the methods that will be used

% The within-trial economic analysis will be performed using
% individual patient level data from the [trial name]. The analytical
% approaches will take the form of cost-effectiveness and cost-utility
% analysis. Based on trial evidence, incremental cost-effectiveness
% (and cost-utility) ratios will be calculated by taking a ratio of
% the difference in the mean costs and mean effects (or utility
% measure).

\subsection{Jurisdiction(s)}

Specify the jurisdiction(s) in which the analysis will be conducted
including details of the country(s) and health system(s)

% The trial is conducted in the UK which has a national health service
% (NHS), providing publicly funded healthcare, primarily free of
% charge at the point of use.


\subsection{Perspective(s)}

State the perspective(s) from which the economic analysis is being
conducted, such as societal perspective and/or healthcare payer
perspective

% The primary economic analysis will be from the NHS and personal
% social services (PSS) perspective. A secondary analysis will include
% the perspective of patients and carers.


\subsection{Time horizon(s)}

State the time horizon(s) over which costs and consequences are being evaluated

% The primary economic analysis will compare the costs and
% consequences of each arm over the first 26 weeks after
% randomisation. A secondary analysis will extend this to compare
% costs and benefits over the two -year follow-up period from
% randomisation.


\section{Economic data collection \& management}

\subsection{Statistical software}

Specify the statistical software that will be used to carry out the
health economic analysis

% Stata version x.x or higher will be used for exploratory analysis
% and Stata and/or MLWin for the main statistical analysis involving
% multilevel multivariable regression.


\subsection{Identification of resources}

Justify and describe items of resource use that will be measured as
part of the trial

% The following items of health care resource use that may differ
% between arms will be measured: health service resource use,
% productivity losses, personal expenditure on health care.  In
% addition staff training and delivery of the intervention will be
% measured to assess the cost of the intervention.


\subsection{Measurement of resource-use data}

Describe the resource-use data collection method(s) (including
external routine datasets) and the time points at which they will be
used.

% Resource-use data will be collected until 18 months post
% randomisation using trial case report forms (CRFs), completed by the
% trial nurses. Patient questionnaires (either postal or online) at 6
% and 12 months follow up will be used to collect information on other
% hospital use and community care use.

\subsection{Valuation of resource-use data}

For each resource item measured, describe how the unit cost will be
derived and from which specific price year. Outline how adjustments
will be made for sources from different price years and which
inflation index will be used.

% All resource use will be valued in monetary terms using appropriate
% UK unit costs or participant valuations estimated at the time of
% analysis (2019-2020). Adjustments will be made for inflation using
% the ONS GDP deflator index.  NHS reference costs will be employed to
% value hospital resource use (e.g. A&E visits and outpatient
% attendances). Medication costs will be taken from the British
% National Formulary (BNF) and the Prescription Cost Analysis (PCA)
% for England.

\subsection{Identification of outcome(s)}

Specify and justify the outcome(s) that will be measured

% The primary economic outcome measure will be Quality-Adjusted Life
% Years (QALYs) derived from utility scores, obtained using the
% EQ-5D-5L quality of life instrument.


\subsection{Measurement of outcome(s)}

Describe the outcome data collection method(s) and the time points at
which they will be used

% Measurements will be recorded prior to randomisation at baseline,
% and at 2 weeks, 6 weeks and 12 weeks post randomisation. Baseline
% and research follow-up assessments will take place at the
% participant's home, general practice or at university premises.

\subsection{Valuation of outcome(s)}

For each outcome measured, describe how it will be valued and the source of these valuations

% Utility scores will be derived from responses to the EQ-5D-5L.  UK
% utility values will be derived using the approach recommended by
% NICE, which is currently using the validated mapping function from
% the existing EQ-5D (-3L). These will be used to form QALYs over the
% 12-week period, adjusting for any imbalances in baseline EQ-5D-5L
% scores.

\subsection{Monitoring collection of health economic data (OPTIONAL)}

Outline how the health economic data collected will be monitored

% Training will be provided to individuals responsible for
% administering the health economics questionnaires. The trial health
% economist(s) will work closely with the trial team throughout the
% data collection period. Data collection forms will be assessed
% throughout the trial period to monitor quality of the data and amend
% any forms or procedures if necessary.


\subsection{Database management (OPTIONAL)}

Outline how the economic data will be stored and managed and by whom

% Economic data will be securely stored on the trial database and
% managed by the trial database manager, [name(s)].


\subsection{Data entry (OPTIONAL)}

Outline how data will be entered/handled and outline any checking systems in place

% All baseline data will be entered into the case report form (CRF) by
% the trial research nurses at the recruitment site. Follow-up data
% collected from postal questionnaires will be entered by the central
% research team.  The database will use controls to limit data entry
% to plausible values.


\subsection{Data archiving (OPTIONAL)}

State whether datasets, interim datasets and final analysis will be archived, and if so, how

% A copy of health economic analysis files, derived datasets, interim
% datasets and final analysis will be locked and archived. Archived
% datasets will be held by [organisation name] and will conform to the
% department data security policy and department data compliance and
% Data Protection Act policies.




\section{Economic data analysis}


\subsection{Analysis population}

Outline the analysis population that will be used in the economic
base-case analysis (such as intention to treat, per protocol)

% The full analysis set will include all randomised participants,
% which is in accordance with the “intention to treat” (ITT)
% principle. A per protocol set will include all participants in the
% full analysis set who are deemed to have no major protocol
% violations (e.g. patient not receiving any of the intended
% intervention).


\subsection{Timing of analyses}

Describe the timing of all planned analyses (e.g. interim and final
analyses)

% The primary (“final”) analysis will be conducted once all patients
% have been followed for two years after the first dose of [trial
% drug], although an interim analysis will be conducted on year 1 data
% once all patients have been followed for one year. The interim
% analysis will take a one-year time horizon and use only data
% collected in patients’ first year of follow-up, with no
% extrapolation. The final analysis will include a within-trial
% analysis, taking a two-year time horizon and extrapolating beyond
% the end of the trial.

\subsection{Discount rates for costs and benefits}

Detail the source of, and justification for, discount rates used for
costs and benefits

% Costs and benefits will be discounted at 3.5\% p.a. as recommended by NICE.


\subsection{Cost-effectiveness threshold(s)}

Detail the cost-effectiveness threshold(s) to be used in
analysis/interpretation

% The estimated mean QALYs and costs associated with each treatment
% option will be combined with a feasible range of values for decision
% makers’ willingness-to-pay (ʎ), to obtain the distribution of net
% benefits at different levels of ʎ. The primary economic analysis
% will use a cost-effectiveness threshold of £20,000 per QALY.

\subsection{Statistical decision rule(s)}

Describe how inference will be drawn (e.g. significance level,
confidence intervals or mean net benefit)

% Mean differences in costs, QALYs and net benefits between the
% treatment groups will be estimated with associated
% 95% confidence intervals.


\subsection{Analysis of resource use}

Describe how differences in the use of resources/services between
randomised groups will be compared

% Differences in the use of services between randomised groups will be
% described but not compared statistically.

\subsection{Analysis of costs}

Describe analyses of the cost data, specifying any covariates for
statistical adjustment, assumptions, and alternative methods

% Differences in overall mean costs between the arms will be analysed
% initially using Ordinary Least Squares (OLS) regression adjusting
% for the minimisation variables of the randomisation procedure. The
% distribution of residuals from the regression model will then be
% examined and a decision will be made as to whether OLS is
% appropriate or another type of regression model should be considered
% (e.g. Generalised Linear Models (GLM)).


\subsection{Analysis of outcomes}

For each outcome used in the economic analysis, describe how the
outcome will be analysed, specifying any covariates for statistical
adjustment, assumptions, and alternative methods.

% An appropriate regression model will be used to adjust for any
% imbalance in baseline utility and the minimisation variables of the
% randomisation process.


\subsection{Data cleaning for analysis}

Outline how data will be cleaned before analysis

% Face validity tests will be conducted on data (e.g. to identify
% misspelt text) and checked against the source documents. Corrections
% made will be documented in the Stata code.

\subsection{Missing data}

Specify the procedure for dealing with missing data

% Trial data will be examined for any missing data. The appropriate
% method for dealing with missing data will depend on the proportion
% of missing data and likely mechanism of missingness. For example,
% multiple imputation methods may be used if the data is missing at
% random (MAR).


\subsection{Analysis of cost-effectiveness}

Describe the methods that will be used to summarise
cost-effectiveness.

% Cost and QALY data will be combined to calculate an incremental
% cost-effectiveness ratio (ICER) and net monetary benefit (NMB)
% statistic from the NHS and PSS perspective. Seemingly Unrelated
% Regression (SUR) will be used, if appropriate, to account for the
% correlation between the costs and the QALYS.


\subsection{Sampling uncertainty}

Describe how uncertainty around the costs and effectiveness estimates
and summary cost-effectiveness measures will be explored

% The nonparametric bootstrapping approach will be used to determine the
% level of sampling uncertainty surrounding the mean ICER by generating
% 10,000 estimates of incremental costs and benefits.


\subsection{Subgroup analyses or analysis of heterogeneity}

Describe any analyses of subgroups or heterogeneity in cost-effectiveness and the analysis methods used

% Analyses will also be conducted on the final dataset to investigate
% how cost-effectiveness varies between different patient subgroups
% (e.g. age <60/60+; sex). Any subgroup analyses for which the smaller
% subgroup includes fewer than 50 participants will be omitted.


\subsection{Sensitivity analyses}

Describe any sensitivity analyses and their form

% Several sensitivity analyses will be undertaken to explore
% uncertainties surrounding key parameters in the economic
% evaluation. The results for complete cost and quality of life data
% (i.e. those with no missing data) as well as a strict per-protocol
% analysis of the data will be provided to identify the impact of
% missing data on the analysis and any sensitivity to protocol
% violations.

\section{Modelling}

\subsection{Extrapolation or decision analytic modelling}

Outline whether decision analytic modelling or any other extrapolation
will be used to estimate cost-effectiveness results beyond the period
of the trial or to introduce an additional comparator or other
evidence.

% Decision analytic modelling will be undertaken (e.g. to extrapolate
% costs & outcomes beyond the follow up period of trial), irrespective
% of statistical significance in trial results, if there is potential
% for the cost -effectiveness of active interventions to improve under
% a longer analysis time horizon than the 18 months follow-up and /or
% under a more efficient risk screening process.


\subsection{Model type}

Describe the modelling approach that will be used and duration of
extrapolation

% A long-term Markov decision model with annual cycle will be used to
% evaluate effects of intervention on costs, health gains and
% cost-effectiveness over the patient’s lifetime.

\subsection{Model structure}

Detail the model structure (where possible, include diagram of model states and transitions between them)

% The decision model will comprise three arms, with one arm for each
% intervention evaluated. Each arm will be structured as a Markov
% model built around health states to which health care cost and
% health related quality of life (HRQoL) data collected as part of the
% trial will be linked. The model structure is shown in Figure X.


\subsection{Treatment effect beyond the end of the trial}

Describe the duration and size of treatment effect in the period
beyond the end of the trial

% The hazard ratio for disease specific mortality observed during the
% trial will be assumed to persist after the end of trial follow up.

\subsection{Other key assumptions}

List the key structural assumptions of the model

% The main structural assumptions underpinning the decision model are
% listed below [list those that are relevant to specific
% disease/trial]. The model structure maybe subject to change,
% following initial exploratory analysis of trial data.

\subsection{Methods for identifying and estimating parameters}

For each model parameter, describe the methods and data sources that
will be used to estimate the parameter (e.g. from the RCT, systematic
review, meta-analysis, other published data or expert opinion)

% The model will require 4 main sets of parameters, which are expected
% to be mainly derived from the trial data. 1) Transition
% probabilities between health states, 2) Treatment Effect/s of the
% intervention, 3) Quality of life decrements, 4) Health care
% costs. Transition probabilities between health states will be
% estimated by sex and 5 or 10-year age groups using trial
% data. Quality of life decrements and costs associated with a fall or
% fracture will be derived from patient questionnaires, analysed using
% regression analysis and routine hospital admissions data.  Other
% parameters, for example related to ‘natural history of disease’,
% will be parameterised using published data, and where necessary
% using formally elicited expert opinion.

\subsection{Model uncertainty}

Describe the methods that will be used to assess parameter uncertainty
in the results. Describe sensitivity analyses for the impact of other
types of uncertainty on results.

% Parameter uncertainty will be assessed using probabilistic
% sensitivity analysis (i.e. by fitting a probability distribution to
% each uncertain parameter) and running Monte Carlo simulations


\subsection{Model validation}

Describe the methods and data that will be used to check the face,
internal and external validity of the model

% The model will be estimated and internally validated using trial
% data.

\subsection{Subgroup analyses/heterogeneity}

Describe subgroup or heterogeneity analyses that will be executed and
reported within the extrapolation or decision analytic modelling

% The model will be used to evaluate the cost-effectiveness of the
% intervention in categories of trial participants by age (50-59;
% 60-69; 70+) and estimated 5-year vascular disease risk
% (<10%; ≥10,<20%; ≥20%).


\subsection{Value of information analysis (OPTIONAL)}

Describe whether value of information analysis is planned and the type
and methods that will be used to calculate value of information

% As part of the analysis we will conduct a value of information
% analysis (VOI). Expected value of perfect information (EVPI) will be
% used to indicate whether the research is potentially worthwhile. In
% addition, the expected value of perfect information for parameters
% (EVPPI) will also be calculated to identify those parameters for
% which more precise estimates would be most valuable.



\section{Reporting/publishing}

\subsection{Reporting standards}

Describe any guidelines that will be followed when publishing results

% CHEERS guidelines will be followed when reporting the health
% economic evaluation, in a format appropriate to stakeholders and
% policy makers.

\subsection{Deviations from the HEAP}

Describe the procedure for reporting any deviations from the HEAP

% Any deviation from HEAP will be described and justified in the final
% published report.

\section{Appendices}


\subsection{Health economic collection tools}

Include template examples of the resource-use data collection sheets
and resource-use questionnaires

% Data collection sheets on hospital stays, visits to A&E, and home
% visits are given in Appendix X and include health utility
% measures. The resource-use questionnaire has been deposited in DIRUM
% (http://www.dirum.org/).

\subsection{Cross-referencing to other trial documents (OPTIONAL)}

Reference to other relevant trial documents that are adhered to and
followed when writing the HEAP and any other references used when
writing the HEAP

% The [name of trials unit] Standard Operating Procedure for Economic
% Evaluations (version 2.1) was followed in designing this analysis.
% The analysis described in this plan adheres to the [organisation
% name] Data Management Plan (version 1.1).


\subsection{Illustrations (OPTIONAL)}

Illustrations such as annotated questionnaires detailing the database
fieldnames, flow charts outlining the flow of data for the economic
evaluation, or template tables

% The conduct and procedures for the economic evaluation will be fully
% integrated into the [name] trial from planning and designing the
% economic evaluation through to publication. The following flowchart
% should be applied.


"""

spirit = r"""

Nella compilazione delle parti di propria competenza si consiglia di
fare riferimento al documento
\href{https://www.ncbi.nlm.nih.gov/pmc/articles/pmid/23303884/}{SPIRIT
  2013 explanation and elaboration}.


\section{Protocol}

% --------------------------------------------------------------
\subsection{Administrative information}

\subsubsection{Title}

Item 1: Descriptive title identifying the study design, population,
interventions, and, if applicable, trial acronym

\subsubsection{Trial registration}

\begin{itemize}
\item Item 2a: Trial identifier and registry name. If not yet registered, name
  of intended registry
\item Item 2b: All items from the World Health Organization Trial Registration
  Data Set
\end{itemize}

\subsubsection{Protocol version}

Item 3: Date and version identifier

\subsubsection{Funding}

Item 4: Sources and types of financial, material, and other support

\subsubsection{Roles and responsibilities}

\begin{itemize}
\item Item 5a: Names, affiliations, and roles of protocol contributors
\item Item 5b:   Name and contact information for the trial sponsor
\item Item 5c: Role of study sponsor and funders, if any, in study design;
  collection, management, analysis, and interpretation of data;
  writing of the report; and the decision to submit the report for
  publication, including whether they will have ultimate authority
  over any of these activities
\item Item 5d: Composition, roles, and responsibilities of the coordinating
  centre, steering committee, endpoint adjudication committee, data
  management team, and other individuals or groups overseeing the
  trial, if applicable (see Item 21a for data monitoring committee)
\end{itemize}


% --------------------------------------------------------------
\subsection{Introduction}

\subsubsection{Background and rationale}

\begin{itemize}
\item Item 6a:  Description of research question and justification for
  undertaking the trial, including summary of relevant studies
  (published and unpublished) examining benefits and harms for each
  intervention
\item Item 6b: Explanation for choice of comparators
\end{itemize}


\subsubsection{Objectives}

Item 7: Specific objectives or hypotheses

\subsubsection{Trial design}

Item 8: Description of trial design including type of trial (eg, parallel
group, crossover, factorial, single group), allocation ratio, and
framework (eg, superiority, equivalence, noninferiority, exploratory)

% ------------------------------------------------------------------
\subsection{Methods: Participants, interventions, and outcomes}

\subsubsection{Study setting}

Item 9: Description of study settings (eg, community clinic, academic
hospital) and list of countries where data will be
collected. Reference to where list of study sites can be obtained

\subsubsection{Eligibility criteria}

Item 10: Inclusion and exclusion criteria for participants. If applicable,
eligibility criteria for study centres and individuals who will
perform the interventions (eg, surgeons, psychotherapists)


\subsubsection{Interventions}

\begin{itemize}
\item Item 11a: Interventions for each group with sufficient detail to allow
  replication, including how and when they will be administered
\item Item 11b: Criteria for discontinuing or modifying allocated interventions
  for a given trial participant (eg, drug dose change in response to
  harms, participant request, or improving/worsening disease)
\item Item 11c: Strategies to improve adherence to intervention protocols, and
  any procedures for monitoring adherence (eg, drug tablet return,
  laboratory tests)
\item Item 11d: Relevant concomitant care and interventions that are permitted
  or prohibited during the trial
\end{itemize}

\subsubsection{Outcomes}

Item 12: Primary, secondary, and other outcomes, including the specific
measurement variable (eg, systolic blood pressure), analysis metric
(eg, change from baseline, final value, time to event), method of
aggregation (eg, median, proportion), and time point for each
outcome. Explanation of the clinical relevance of chosen efficacy and
harm outcomes is strongly recommended

\subsubsection{Participant timeline}

Item 13: Time schedule of enrolment, interventions (including any run-ins and
washouts), assessments, and visits for participants. A schematic
diagram is highly recommended (see Figure)

\subsubsection{Sample size}

Item 14: Estimated number of participants needed to achieve study objectives
and how it was determined, including clinical and statistical
assumptions supporting any sample size calculations

\subsubsection{Recruitment}

Item 15: Strategies for achieving adequate participant enrolment to reach
target sample size

% -----------------------------------------------------------------------
\subsection{Methods: Assignment of interventions (for controlled trials)}

\subsubsection{Allocation sequence generation}

Item 16a: Method of generating the allocation sequence (eg, computer-generated
random numbers), and list of any factors for stratification. \\
To reduce predictability of a random sequence, details of any planned
restriction (eg, blocking) should be provided in a separate document
that is unavailable to those who enrol participants or assign
interventions

\subsubsection{Allocation concealment mechanism}

Item 16b: Mechanism of implementing the allocation sequence (eg, central
telephone; sequentially numbered, opaque, sealed envelopes),
describing any steps to conceal the sequence until interventions are
assigned

\subsubsection{Allocation implementation}

Item 16c: Who will generate the allocation sequence, who will enrol
participants, and who will assign participants to interventions

\subsubsection{Blinding (masking)}

\begin{itemize}
\item Item 17a: Who will be blinded after assignment to interventions
  (eg, trial participants, care providers, outcome assessors, data
  analysts), and how
\item Item 17b: If blinded, circumstances under which unblinding is
  permissible, and procedure for revealing a participant’s allocated
  intervention during the trial
\end{itemize}


% -----------------------------------------------------------------------
\subsection{Methods: Data collection, management, and analysis}

\subsubsection{Data collection methods}

\begin{itemize}
\item Item 18a: Plans for assessment and collection of outcome,
  baseline, and other trial data, including any related processes to
  promote data quality (eg, duplicate measurements, training of
  assessors) and a description of study instruments (eg,
  questionnaires, laboratory tests) along with their reliability and
  validity, if known. Reference to where data collection forms can be
  found, if not in the protocol
\item Item 18b: Plans to promote participant retention and complete
  follow-up, including list of any outcome data to be collected for
  participants who discontinue or deviate from intervention protocols
\end{itemize}


\subsubsection{Data management}

Item 19: Plans for data entry, coding, security, and storage, including any
related processes to promote data quality (eg, double data entry;
range checks for data values). Reference to where details of data
management procedures can be found, if not in the protocol


\subsubsection{Statistical methods}
\begin{itemize}
\item Item 20a: Statistical methods for analysing primary and secondary
  outcomes. Reference to where other details of the statistical
  analysis plan can be found, if not in the protocol
\item Item 20b: Methods for any additional analyses (eg, subgroup and adjusted
  analyses)
\item Item 20c: Definition of analysis population relating to protocol
  non-adherence (eg, as randomised analysis), and any statistical
  methods to handle missing data (eg, multiple imputation)
\end{itemize}

% -----------------------------------------------------------------------
\subsection{Methods: Monitoring}

\subsubsection{Data monitoring}

\begin{itemize}
\item Item 21a: Composition of data monitoring committee (DMC); summary of its
  role and reporting structure; statement of whether it is independent
  from the sponsor and competing interests; and reference to where
  further details about its charter can be found, if not in the
  protocol. Alternatively, an explanation of why a DMC is not needed
\item Item 21b: Description of any interim analyses and stopping guidelines,
  including who will have access to these interim results and make the
  final decision to terminate the trial
\end{itemize}

\subsubsection{Harms}

Item 22: Plans for collecting, assessing, reporting, and managing
solicited and spontaneously reported adverse events and other
unintended effects of trial interventions or trial conduct


\subsubsection{Auditing}
Item 23: Frequency and procedures for auditing trial conduct, if any,
and whether the process will be independent from investigators and the
sponsor


% -----------------------------------------------------------------------
\subsection{Ethics and dissemination}

\subsubsection{Research ethics approval}

Item 24: Plans for seeking research ethics committee/institutional
review board (REC/IRB) approval

\subsubsection{Protocol amendments}

Item 25: Plans for communicating important protocol modifications (eg,
changes to eligibility criteria, outcomes, analyses) to relevant
parties (eg, investigators, REC/IRBs, trial participants, trial
registries, journals, regulators)

\subsubsection{Consent or assent}

\begin{itemize}
\item Item 26a: Who will obtain informed consent or assent from
  potential trial participants or authorised surrogates, and how (see
  Item 32)
\item Item 26b: Additional consent provisions for collection and use of
  participant data and biological specimens in ancillary studies, if
  applicable
\end{itemize}

\subsubsection{Confidentiality}

Item 27: How personal information about potential and enrolled participants
will be collected, shared, and maintained in order to protect
confidentiality before, during, and after the trial

\subsubsection{Declaration of interests}

Item 28: Financial and other competing interests for principal investigators
for the overall trial and each study site

\subsubsection{Access to data}

Item 29: Statement of who will have access to the final trial dataset, and
disclosure of contractual agreements that limit such access for
investigators

\subsubsection{Ancillary and post-trial care}

Item 30: Provisions, if any, for ancillary and post-trial care, and for
compensation to those who suffer harm from trial participation

\subsubsection{Dissemination policy}

\begin{itemize}
\item Item 31a: Plans for investigators and sponsor to communicate
  trial results to participants, healthcare professionals, the public,
  and other relevant groups (eg, via publication, reporting in results
  databases, or other data sharing arrangements), including any
  publication restrictions
\item Item 31b: Authorship eligibility guidelines and any intended use of
  professional writers
\item Item 31c: Plans, if any, for granting public access to the full
  protocol, participant-level dataset, and statistical code
\end{itemize}
% -----------------------------------------------------------------------
\subsection{Appendices}


\subsubsection{Informed consent materials}

Item 32: Model consent form and other related documentation given to
participants and authorised surrogates

\subsubsection{Biological specimens}

Item 33: Plans for collection, laboratory evaluation, and storage of
biological specimens for genetic or molecular analysis in the current
trial and for future use in ancillary studies, if applicable


"""

trials = r"""
% STARTHERE

Template utilizzato da Trials, disponibile
\href{https://trialsjournal.biomedcentral.com/submission-guidelines/preparing-your-manuscript/study-protocol/structured-study-protocol-template}{qui},
che riorganizza gli item dello
\href{http://www.equator-network.org/reporting-guidelines/spirit-2013-statement-defining-standard-protocol-items-for-clinical-trials}{SPIRIT}
per rendere il tutto più completo/scorrevole.  Le indicazioni in
italico sono divise in Trials guidance (fonte: detto giornale ed
obbligatorie per esso) o Spirit guidance (fonte SPIRIT: meglio
considerarle per qualsiasi journal): tutto il testo in italico si può
cancellare dopo aver compilato. Se un item non si applica fornire una
spiegazione, non cancellarlo dal template.  I numeri tra graffe
indicano gli item di SPIRIT sono da lasciare.

Nella compilazione delle parti di propria competenza si consiglia di
fare riferimento al documento
\href{https://www.ncbi.nlm.nih.gov/pmc/articles/pmid/23303884/}{SPIRIT
  2013 explanation and elaboration}.


\section{Title}

{\em SPIRIT guidance: Descriptive title identifying the study design,
population, interventions, and, if applicable, trial acronym.}

\section{Protocol contributors}

{\em Trials guidance: Names of protocol contributors.}

\section{Abstract}

{\em Trials guidance: The Abstract should not exceed 350 words. Please
minimize the use of abbreviations and do not cite references in the
abstract. The abstract must include the following separate sections:
\begin{itemize}
\item
  \textbf{Background}: the context and purpose of the study
\item
  \textbf{Methods}: how the study will be performed
\item
  \textbf{Discussion}: a brief summary and potential implications
\item \textbf{Trial registration}: If your article reports the results
  of a health care intervention on human participants, it must be
  registered in an appropriate registry and the registration number
  and date of registration should be in stated in this section. If it
  was not registered prospectively (before enrollment of the first
  participant), you should include the words 'retrospectively
  registered'. See our editorial policies for more information on
  trial registration.
\end{itemize}}

\section{Keywords}

{\em Trials guidance: Three to ten keywords representing the main
content of the article.}

\section{Administrative information}

{\em Trials guidance: please include this text in your protocol just
above the Administrative information table:}

Note: the numbers in curly brackets in this protocol refer to SPIRIT
checklist item numbers. The order of the items has been modified to
group similar items (see
\url{http://www.equator-network.org/reporting-guidelines/spirit-2013-statement-defining-standard-protocol-items-for-clinical-trials/}).


\begin{table}
  \centering
  \begin{tabular}{|l|l|}
    \hline
    Title \{1\} &
                  {\em SPIRIT guidance: Descriptive title identifying
                  the study design, population, interventions, and, if
                  applicable, trial acronym. }\\
    \hline
    Trial registration \{2a and 2b\}
                &
                  {\em SPIRIT guidance: Trial identifier and registry name.
                  If not yet registered, name of intended registry.

                  Item 2b is met if the register used for registration
                  collects all items from the World Health Organization
                  Trial Registration Data Set.}\\
    \hline
    Protocol version \{3\}
                &
                  {\em SPIRIT guidance: Date and version identifier.}\\
    \hline
    Funding \{4\}
                &
                  {\em  SPIRIT guidance: Sources and types of financial,
                  material, and other support.}\\
    \hline
    Author details \{5a\}
                &
                  {\em SPIRIT guidance: Affiliations of protocol
                  contributors.}\\
    \hline
    Name and contact information for the trial sponsor \{5b\}
                &
                  {\em SPIRIT guidance: Name and contact information for
                  the trial sponsor.}\\                  
    \hline
    Role of sponsor \{5c\}
                & {\em
                  SPIRIT guidance: Role of study sponsor and
                  funders, if any, in study design; collection,
                  management, analysis, and
                  interpretation of data; writing of the report;
                  and the decision to
                  submit the report for publication, including whether they
                  will have ultimate authority over any of these activities.}\\
    \hline
  \end{tabular}
\end{table}



\section{Introduction}

\subsection{Background and rationale \{6a\}}

{\em
SPIRIT guidance: Description of research question and justification for
undertaking the trial, including summary of relevant studies (published
and unpublished) examining benefits and harms for each intervention.
}

\subsection{Objectives \{7\}}

{\em
SPIRIT guidance: Specific objectives or hypotheses.
}

\subsection{Trial design \{8\}}

{\em
SPIRIT guidance: Description of trial design including type of trial
(eg, parallel group, crossover, factorial, single group), allocation
ratio, and framework (eg, superiority, equivalence, noninferiority,
exploratory).
}
\section{Methods: Participants, interventions and outcomes}

\subsection{Study setting \{9\}}

{\em
SPIRIT guidance: Description of study settings (eg, community clinic,
academic hospital) and list of countries where data will be collected.
Reference to where list of study sites can be obtained.
}

\subsection{Eligibility criteria \{10\}}

{\em
SPIRIT guidance: Inclusion and exclusion criteria for participants. If
applicable, eligibility criteria for study centres and individuals who
will perform the interventions (eg, surgeons, psychotherapists).
}

\subsection{Who will take informed consent? \{26a\}}

{\em
SPIRIT guidance: Who will obtain informed consent or assent from
potential trial participants or authorised surrogates, and how (see Item
32).
}

\subsection{Additional consent provisions for collection and use of
  participant data and biological specimens \{26b\}}

{\em
SPIRIT guidance: Additional consent provisions for collection and use of
participant data and biological specimens in ancillary studies, if
applicable.
}

\section{Interventions}

\subsection{Explanation for the choice of comparators \{6b\}}

{\em
SPIRIT guidance: Explanation for choice of comparators.
}

\subsection{Intervention description \{11a\}}

{\em
SPIRIT guidance: Interventions for each group with sufficient detail to
allow replication, including how and when they will be administered.
}

\subsection{Criteria for discontinuing or modifying allocated
  interventions \{11b\}}

{\em
SPIRIT guidance: Criteria for discontinuing or modifying allocated
interventions for a given trial participant (eg, drug dose change in
response to harms, participant request, or improving/worsening disease).
}

\subsection{Strategies to improve adherence to interventions \{11c\}}

{\em
SPIRIT guidance: Strategies to improve adherence to intervention
protocols, and any procedures for monitoring adherence (eg, drug tablet
return, laboratory tests).
}

\subsection{Relevant concomitant care permitted or prohibited during the
trial \{11d\}}

{\em
SPIRIT guidance: Relevant concomitant care and interventions that are
permitted or prohibited during the trial.
}

\subsection{Provisions for post-trial care \{30\}}

{\em
SPIRIT guidance: Provisions, if any, for ancillary and post-trial care,
and for compensation to those who suffer harm from trial participation.
}

\subsection{Outcomes \{12\}}

{\em
SPIRIT guidance: Primary, secondary, and other outcomes, including the
specific measurement variable (eg, systolic blood pressure), analysis
metric (eg, change from baseline, final value, time to event), method of
aggregation (eg, median, proportion), and time point for each outcome.
Explanation of the clinical relevance of chosen efficacy and harm
outcomes is strongly recommended.
}

\subsection{Participant timeline \{13\}}

{\em
SPIRIT guidance: Time schedule of enrolment, interventions (including
any run-ins and washouts), assessments, and visits for participants. A
schematic diagram is highly recommended (see figure at
\href{http://www.spirit-statement.org/publications-downloads/}{{http://www.spirit-statement.org/publications-downloads/}}).
}

\subsection{Sample size \{14\}}

{\em
SPIRIT guidance: Estimated number of participants needed to achieve
study objectives and how it was determined, including clinical and
statistical assumptions supporting any sample size calculations.
}

\subsection{Recruitment \{15\}}

{\em
SPIRIT guidance: Strategies for achieving adequate participant enrolment
to reach target sample size.
}

\section{Assignment of interventions: allocation}

\subsection{Sequence generation \{16a\}}

{\em
SPIRIT guidance: Method of generating the allocation sequence (eg,
computer-generated random numbers), and list of any factors for
stratification. To reduce predictability of a random sequence, details
of any planned restriction (eg, blocking) should be provided in a
separate document that is unavailable to those who enrol participants or
assign interventions.
}

\subsection{Concealment mechanism \{16b\}}

{\em
SPIRIT guidance: Mechanism of implementing the allocation sequence (eg,
central telephone; sequentially numbered, opaque, sealed envelopes),
describing any steps to conceal the sequence until interventions are
assigned.
}

\subsection{Implementation \{16c\}}

{\em
SPIRIT guidance: Who will generate the allocation sequence, who will
enrol participants, and who will assign participants to interventions.
}

\section{Assignment of interventions: Blinding}

\subsection{Who will be blinded \{17a\}}

{\em
SPIRIT guidance: Who will be blinded after assignment to interventions
(eg, trial participants, care providers, outcome assessors, data
analysts), and how.
}

\subsection{Procedure for unblinding if needed \{17b\}}

{\em
SPIRIT guidance: If blinded, circumstances under which unblinding is
permissible, and procedure for revealing a participant's allocated
intervention during the trial.
}

\section{Data collection and management}

\subsection{Plans for assessment and collection of outcomes \{18a\}}

{\em
SPIRIT guidance: Plans for assessment and collection of outcome,
baseline, and other trial data, including any related processes to
promote data quality (eg, duplicate measurements, training of assessors)
and a description of study instruments (eg, questionnaires, laboratory
tests) along with their reliability and validity, if known. Reference to
where data collection forms can be found, if not in the protocol.
}

\subsection{Plans to promote participant retention and complete follow-up
\{18b\}}

{\em
SPIRIT guidance: Plans to promote participant retention and complete
follow-up, including list of any outcome data to be collected for
participants who discontinue or deviate from intervention protocols.
}
\subsection{Data management \{19\}}

{\em
SPIRIT guidance: Plans for data entry, coding, security, and storage,
including any related processes to promote data quality (eg, double data
entry; range checks for data values). Reference to where details of data
management procedures can be found, if not in the protocol.
}
\subsection{Confidentiality \{27\}}

{\em
SPIRIT guidance: How personal information about potential and enrolled
participants will be collected, shared, and maintained in order to
protect confidentiality before, during, and after the trial.
}
\subsection{Plans for collection, laboratory evaluation and storage of
biological specimens for genetic or molecular analysis in this
trial/future use \{33\}}

{\em
SPIRIT guidance: Plans for collection, laboratory evaluation, and
storage of biological specimens for genetic or molecular analysis in the
current trial and for future use in ancillary studies, if applicable.
}
\section{Statistical methods}

\subsection{Statistical methods for primary and secondary outcomes \{20a\}}

{\em
SPIRIT guidance: Statistical methods for analysing primary and secondary
outcomes. Reference to where other details of the statistical analysis
plan can be found, if not in the protocol.
}
\subsection{Interim analyses \{21b\}}

{\em
SPIRIT guidance: Description of any interim analyses and stopping
guidelines, including who will have access to these interim results and
make the final decision to terminate the trial.
}
\subsection{Methods for additional analyses (e.g. subgroup analyses)
\{20b\}}

{\em
SPIRIT guidance: Methods for any additional analyses (eg, subgroup and
adjusted analyses).
}

\subsection{Methods in analysis to handle protocol non-adherence and any
statistical methods to handle missing data \{20c\}}

{\em
SPIRIT guidance: Definition of analysis population relating to protocol
non-adherence (eg, as randomised analysis), and any statistical methods
to handle missing data (eg, multiple imputation).
}

\subsection{Plans to give access to the full protocol, participant
level-data and statistical code \{31c\}}

{\em
SPIRIT guidance: Plans, if any, for granting public access to the full
protocol, participant-level dataset, and statistical code.
}

\section{Oversight and monitoring}

\subsection{Composition of the coordinating centre and trial steering
committee \{5d\}}

{\em
Trials guidance: Provide information on the composition, roles and
responsibilities of the coordinating centre and trial steering committee
and all groups providing day to day support for the trial. There will
always be a group running the trial day-to-day and providing
organisational support and knowing how often they will meet, plus
information on other committees providing oversight such as a Trial
Steering Committee, and how often they will meet over the course of the
trial, is what we need for item 5d. We do not need names of staff.

SPIRIT guidance: Composition, roles, and responsibilities of the
coordinating centre, steering committee, endpoint adjudication
committee, data management team, and other individuals or groups
overseeing the trial, if applicable (see Item 21a for data monitoring
committee).
}

\subsection{Composition of the data monitoring committee, its role and
reporting structure \{21a\}}

{\em
SPIRIT guidance: Composition of data monitoring committee (DMC); summary
of its role and reporting structure; statement of whether it is
independent from the sponsor and competing interests; and reference to
where further details about its charter can be found, if not in the
protocol. Alternatively, an explanation of why a DMC is not needed.
}

\subsection{Adverse event reporting and harms \{22\}}

{\em
SPIRIT guidance: Plans for collecting, assessing, reporting, and
managing solicited and spontaneously reported adverse events and other
unintended effects of trial interventions or trial conduct.
}

\subsection{Frequency and plans for auditing trial conduct \{23\}}

{\em
SPIRIT guidance: Frequency and procedures for auditing trial conduct, if
any, and whether the process will be independent from investigators and
the sponsor.
}

\subsection{Plans for communicating important protocol amendments to
relevant parties (e.g. trial participants, ethical committees) \{25\}}

{\em
SPIRIT guidance: Plans for communicating important protocol
modifications (eg, changes to eligibility criteria, outcomes, analyses)
to relevant parties (eg, investigators, REC/IRBs, trial participants,
trial registries, journals, regulators).
}

\section{Dissemination plans \{31a\}}

{\em SPIRIT guidance: Plans for investigators and sponsor to
  communicate trial results to participants, healthcare professionals,
  the public, and other relevant groups (eg, via publication,
  reporting in results databases, or other data sharing arrangements),
  including any publication restrictions.}

\section{Discussion}

{\em Trials guidance: This should include a discussion of any
  practical or operational issues involved in performing the study and
  any issues not covered in other sections.}

\section{Trial status}

{\em Trials guidance: Authors should report the protocol version
number and date, the date recruitment began, and the approximate date
when recruitment will be completed.}

\section{Abbreviations}

{\em Trials guidance: If abbreviations are used in the text they
should be defined in the text at first use, and a list of abbreviations
should be provided.}

\section{Declarations}

{\em
Trials guidance: All manuscripts must contain the following
subheadings:
\begin{itemize}
\item Acknowledgements
\item Authors' contributions
\item Funding
\item Availability of data and material
\item Ethics approval and consent to participate
\item Consent for publication
\item Competing interests
\item Authors' information (optional)
\end{itemize}
}

\subsection{Acknowledgements}

{\em

\emph{Trials} guidance: Please acknowledge anyone who contributed
towards the article who does not meet the criteria for authorship
including anyone who provided professional writing services or
materials. Authors should obtain permission to acknowledge from all
those mentioned in the Acknowledgements section. See
our \href{http://www.biomedcentral.com/submissions/editorial-policies\#authorship}{editorial
policies} for a full explanation of acknowledgements and authorship
criteria. If you do not have anyone to acknowledge, please write ``Not
applicable'' in this section. 

Group authorship (for manuscripts involving a collaboration group): if
you would like the names of the individual members of a collaboration
Group to be searchable through their individual PubMed records, please
ensure that the title of the collaboration Group is included on the
title page and in the submission system and also include collaborating
author names as the last paragraph of the ``Acknowledgements'' section.
Please add authors in the format First Name, Middle initial(s)
(optional), Last Name. You can add institution or country information
for each author if you wish, but this should be consistent across all
authors. Please note that individual names may not be present in the
PubMed record at the time a published article is initially included in
PubMed as it takes PubMed additional time to code this information.
}

\subsection{Authors' contributions \{31b\}}

{\em SPIRIT guidance: Authorship eligibility guidelines and any
  intended use of professional writers.

Trials guidance: The individual contributions of authors to the
manuscript should be specified in this section. Guidance and criteria
for authorship can be found in
our \href{http://www.biomedcentral.com/submissions/editorial-policies\#authorship}{editorial
policies}. Please use initials to refer to each author's contribution in
this section, for example: ``AB is the Chief Investigator; she conceived
the study, led the proposal and protocol development. CD contributed to
study design and to development of the proposal. EF was the lead trial
methodologist. All authors read and approved the final manuscript.''
}

\subsection{Funding \{4\}}

{\em
SPIRIT guidance: Sources and types of financial, material, and other
support.

Trials guidance: All sources of funding for the research reported
should be declared. You will be required to include a copy of the
original funding document and an English translation of this document as
an additional file on submission, which will be checked against this
declaration. The role of the funding body in the design of the study and
collection, analysis, and interpretation of data and in writing the
manuscript should be declared.
}

\subsection{Availability of data and materials \{29\}}

{\em
SPIRIT guidance: Statement of who will have access to the final trial
dataset, and disclosure of contractual agreements that limit such access
for investigators.

Trials guidance: Please do not include any baseline or pilot data in
your study protocol. The Editorial Office will ask you to remove this
if it is included. Please declare here who will have access to the
final trial dataset and disclose contractual agreements that limit
such access for investigators.
}

\subsection{Ethics approval and consent to participate \{24\}}

{\em
SPIRIT guidance: Plans for seeking research ethics
committee/institutional review board (REC/IRB) approval.

Trials guidance: Trials do not consider study protocols for
studies without ethical approval. You will be required to provide a copy
of the original ethical approval document and an English translation of
this document as an additional file on submission, which will be checked
against this declaration. The name of the ethics committee that approved
the study and the committee's reference number (if applicable) should be
declared. Details of authors' intentions to obtain consent to
participate in the study from participants (or their parent or legal
guardian in the case of children under 16) should be declared. ``eg. ABC
Ethical Review Board ABC123456. Written, informed consent to participate
will be obtained from all participants''
}

\subsection{Consent for publication \{32\}}

{\em
SPIRIT guidance: Model consent form and other related documentation
given to participants and authorised surrogates.

Trials guidance: Please do not include any baseline or pilot data
in your study protocol. The Editorial Office will ask you to remove this
if it is included. If you have included any details, images or videos
relating to an individual person, written informed consent for the
publication of these details must be obtained from that person (or their
parent or legal guardian in the case of children under 18) and declared
in this section. Please also state whether you will be willing to
provide a model consent form on request. If this section does not apply,
please state ``Not applicable''.
}

\subsection{Competing interests \{28\}}

{\em
SPIRIT guidance: Financial and other competing interests for principal
investigators for the overall trial and each study site.

Trials guidance: All financial and non-financial competing
interests must be declared in this section. See
our \href{https://www.biomedcentral.com/getpublished/editorial-policies\#availability+of+data+and+materials}{editorial
policies} for a full explanation of competing interests. If you are
unsure whether you or any of your co-authors have a competing interest
please contact the editorial office. Please use the authors initials to
refer to each authors' competing interests in this section. If you do
not have any competing interests, please state: ``The authors declare
that they have no competing interests'' in this section.
}

\subsection{Authors' information (optional)}

{\em 
Trials guidance: This section is optional.

You may choose to use this section to include any relevant information
about the author(s) that may aid the reader's interpretation of the
article, and understand the standpoint of the author(s). This may
include details about the authors' qualifications, current positions
they hold at institutions or societies, or any other relevant background
information. Please refer to authors using their initials. Note this
section should not be used to describe any competing interests.
}

\section{References}

{\em Seguire il formato richiesto dal giornale dove si vuole pubblicare.}


"""



#!/usr/bin/env python3

import argparse
import os
import subprocess

from pathlib import Path

preamble = r"""
\setcounter{tocdepth}{3}
\setcounter{secnumdepth}{3}
\usepackage[utf8]{inputenc}
\usepackage[english, italian]{babel}
\usepackage{mypkg}
\usepackage{tikz}
\usetikzlibrary{arrows,snakes,backgrounds,calc}
\usepackage{wrapfig}
\usepackage{subfig}
\usepackage{rotating}
\usepackage{cancel}
\usepackage{hyperref}
\usepackage[usefamily=R]{pythontex}
\begin{document}
\tableofcontents
"""

outro = r"\end{document}"


def worker(what):
    argparser = argparse.ArgumentParser()
    argparser.add_argument("path", type=str, help="path to a tex or a dir")
    args = argparser.parse_args()
    inpath = Path(args.path)
    outfname = inpath.stem
    tmp = Path("/tmp")
    tmptex = Path("tmp_%s.tex" % outfname)
    tmppdf = Path("tmp_%s.pdf" % outfname)
    finalpdf = Path("%s.pdf" % outfname)
    if not inpath.exists():
        raise FileNotFoundError
    if inpath.is_dir():
        files = [f for f in list(inpath.iterdir()) if f.suffix == '.tex']
    else:
        files = [inpath]
    contents = []
    for tex in sorted(files):
        with tex.open() as f:
            contents.append(f.read())
    content = "\n".join(contents)
    full_content = (
        r"""\documentclass{%s}""" % what + preamble + content + outro
    )
    with tmptex.open("w") as f:
        f.write(full_content)
    # # here clean if there are old shit in /tmp
    # old_files = tmp / tmp_shit
    # run pdflatex with output in /tmp
    pdflatex_cmd = ["pdflatex", "-output-directory", tmp, tmptex]
    subprocess.run(pdflatex_cmd)
    # # if pythontex code file exists, run pythontex before the next pdflatex
    # pythontexcodef = tmp / tex.with_stem(".pytxcode")
    # if pythontexcodef.exists():
    #     subprocess.run(["pythontex", tmp/tex.stem])
    subprocess.run(pdflatex_cmd)
    # cleaning and renaming (removing "tmp_")
    tmptex.unlink()
    subprocess.run(["mv", tmp / tmppdf, tmp / finalpdf])
    print(
        """All done, open the final file with:

    okular {} &
    """.format(
            tmp / finalpdf
        )
    )


def article():
    worker("article")


def book():
    worker("book")


if __name__ == "__main__":
    book()



"""Import/export handy utilities."""

import os as _os
import pandas as _pd
import tempfile as _tempfile
import zipfile as _zipfile

from .dm import fix_varnames as _fix_varnames

from pathlib import Path as _Path
from typing import Sequence as _Sequence


# ------------------------------------
# Figure and images stuff
# ------------------------------------
def export_figure(
    fig,
    label: str = "",
    caption: str = "",
    fdir: str = "outputs",
    fname: str = "",
    scale: float = 1,
):
    """Dump a figure (save to file, include in LaTeX)

    Save to png, eps and pdf and include in Latex an image;
    intended to be used inside pythontex pycode.

    Args:
       fig (matplotlib.figure.Figure): the fig
       label (str): posta dopo "fig:" in LaTeX
       caption (str): caption LaTeX, se mancante viene riadattata label
       fdir (str): directory dove salvare, se mancante si usa /tmp
       fname (str): basename del file da salvare, se mancante si prova
          a riusare label oppure a creare un file temporaneo
       scale (float): scale di includegraphics di LaTeX

    Returns:
       None: nothing interesting here
    """
    # fdir not existing, using /tmp
    if not _os.path.isdir(fdir):
        fdir = '/tmp'

    # default filename to be set as label, if available or a temporary one
    if fname == "":
        if label != "":
            fname = label
        else:
            tmp = _tempfile.mkstemp(dir=fdir)
            fname = _os.path.basename(tmp[1])

    # produce the file paths
    base_path = _os.path.join(fdir, fname)
    eps_path = base_path + ".eps"
    png_path = base_path + ".png"
    pdf_path = base_path + ".pdf"

    # save figures to hard drive
    fig.savefig(eps_path)
    fig.savefig(png_path)  # , dpi = 600)
    fig.savefig(pdf_path)

    # latex stuff
    latex_label = 'fig:' + label
    caption = (
        label.capitalize().replace("_", " ") if caption == "" else caption
    )
    latex = (
        "\\begin{figure} \\centering "
        + "\\includegraphics[scale=%(scale).2f]{%(base_path)s}"
        + " \\caption{%(caption)s} \\label{%(label)s} \\end{figure}"
    )
    subs = {
        "scale": scale,
        "base_path": base_path,
        "caption": caption,
        "label": latex_label,
    }
    print(latex % subs)


# ------------------------------------
# Dataset Import/export routines
# ------------------------------------
def import_data(fpaths: str | _Path | _Sequence[str | _Path],
                csv_kwargs: dict = {},
                excel_kwargs: dict = {},
                rm_common_prefix: bool = True) -> dict[str, _pd.DataFrame]:
    '''Import data from one or several filepaths (supported formats: .csv
    .xls .xlsx .zip) and return a dict of DataFrame
    '''
    # import ipdb
    # ipdb.set_trace()

    # uniform 1 to many and clean input
    if isinstance(fpaths, str) or isinstance(fpaths, _Path):
        fpaths = [fpaths]
    accepted_fpaths = [
        str(f)
        for f in fpaths
        if _os.path.splitext(f)[1].lower() in {".csv", ".xls", ".xlsx", ".zip"}
    ]

    rval: dict[str, _pd.DataFrame] = {}

    for fpath in accepted_fpaths:
        fname = _os.path.splitext(_os.path.basename(fpath))[0]
        fext = _os.path.splitext(fpath)[1].lower()
        if fext == '.csv':
            dfname = fname
            data = _pd.read_csv(fpath, **csv_kwargs)
            if dfname not in rval.keys():  # check for duplicates
                rval[dfname] = data
            else:
                msg = f"{dfname} is duplicated, skipping to avoid overwriting"
                raise Warning(msg)
        elif fext in {'.xls', '.xlsx'}:
            sheets = _pd.read_excel(
                fpath, None, **excel_kwargs
            )  # import all the sheets as a dict of DataFrame
            sheets = {
                f"{fname}_{k}": v for k, v in sheets.items()
            }  # add xlsx to sheet names
            rval.update(sheets)
        elif (
            fext == '.zip'
        ):  # unzip in temporary directory and go by recursion
            with _tempfile.TemporaryDirectory() as tempdir:
                with _zipfile.ZipFile(fpath) as myzip:
                    myzip.extractall(tempdir)
                    zipped_fpaths = [
                        _os.path.join(tempdir, f) for f in _os.listdir(tempdir)
                    ]
                    zipped_data = import_data(zipped_fpaths)
            # prepend zip name to fname (as keys) and update results
            zipped_data = {
                "f{fname}_{k}": v for k, v in zipped_data.items()
            }
            rval.update(zipped_data)
        else:
            msg = f"File format not supported for {fext}. Ignoring it."
            raise Warning(msg)

    if len(rval) == 1:
        # a single dataset: return the data directly
        for v in rval.values():
            return(v)
    elif len(rval) > 1:
        # multiple dataset return the dict and remove common name prefix
        if rm_common_prefix:
            keys = list(rval.keys())
            common_prefix = _os.path.commonprefix(keys)
            rval = {k.removeprefix(common_prefix) : rval[k] for k in rval.keys()}
        return rval
    else:
        msg = "No data to be imported."
        raise ValueError(msg)



# _rdf: pd.DataFrame to R data.frame converter (a-la dput)
# --------------------------------------------------------
def _rdf_integer(x: _pd.Series, xn: str):
    data_str = x.to_string(
        na_rep="NA",
        index=False,
        header=False,
    ).replace("\n", ", ")
    rval = f"{xn} = c({data_str})"
    return rval

# placeholder
_rdf_numeric = _rdf_integer

def _rdf_factor(x: _pd.Series, xn: str):
    # to be safe it's seems to be better rather than
    # pd.Categorical
    x = _pd.Series(x, dtype='category')
    # raw data (integers)
    data = x.cat.codes
    data_str = "c({})".format(
        data.to_string(
            na_rep="NA",
            index=False,
            header=False,
        ).replace("\n", ", ")
    )
    # unique categories and labels
    categs = x.cat.categories
    levels = []
    labels = []
    for lev, lab in enumerate(categs):
        levels.append(str(lev))
        labels.append("'{}'".format(lab))

    levs = ", ".join(levels)
    labs = ", ".join(labels)
    levels_str = "levels = c({})".format(levs)
    labels_str = "labels = c({})".format(labs)
    # return
    rval = "{} = factor({}, {}, {})".format(
        xn, data_str, levels_str, labels_str
    )
    return rval

def _rdf_object(x: _pd.Series, xn: str):
    data_l = x.to_list()
    data_l2 = []
    for s in data_l:
        if isinstance(s, str) and s != "":
            data_l2.append('"{}"'.format(s))
        else:
            data_l2.append('NA')
    data_str = ', '.join(data_l2)
    rval = "{} = c({})".format(xn, data_str)
    return rval

def _rdf_bool(x: _pd.Series, xn: str):
    ft = {True: "TRUE", False: "FALSE"}
    data_str = (
        x.map(ft)
        .to_string(
            na_rep="NA",
            index=False,
            header=False,
        )
        .replace("\n", ", ")
    )
    rval = "{} = c({})".format(xn, data_str)
    return rval

# Thigs yet TODO
def _rdf_NA(x: _pd.Series, xn: str):
    rval = "{} = NA".format(xn)
    return rval

_rdf_datetime = _rdf_NA

def _rdf(df: _pd.DataFrame, path: str | _Path, dfname: str = "df"):
    """
    pd.DataFrame to R data.frame 'converter'
    """
    path = _Path(path)

    r_code = []
    r_code.append("{} <- data.frame(".format(dfname))
    for var in df.columns:
        x = df[var]
        if _pd.api.types.is_integer_dtype(x):
            r_code.append(_rdf_integer(x, var))
        elif _pd.api.types.is_numeric_dtype(x):
            r_code.append(_rdf_numeric(x, var))
        elif _pd.api.types.is_categorical_dtype(x):
            r_code.append(_rdf_factor(x, var))
        elif _pd.api.types.is_bool_dtype(x):
            r_code.append(_rdf_bool(x, var))
        elif _pd.api.types.is_datetime64_any_dtype(x):
            r_code.append(_rdf_datetime(x, var))
        elif _pd.api.types.is_object_dtype(x):
            r_code.append(_rdf_object(x, var))
        else:
            msg = f"{var}: il tipo {str(x.dtype)} non è ancora gestito."
            raise ValueError(msg)
        is_last = var == df.columns[-1]
        if not is_last:
            r_code.append(",\n")
        else:
            r_code.append(")\n")

    with path.open(mode="w") as f:
        f.writelines(r_code)



def export_data(
        x: _pd.DataFrame | dict[str, _pd.DataFrame],
        path: str | _Path,
        ext: str | list[str] = ["xlsx", "csv", "pkl", "R"],
        index=False
) -> None:
    """Export a DataFrame or a dict of DataFrames as csv/xlsx

    In case of a dict is used and a csv path is given, the path is
    suffixed with dict names

    x: dict of pandas.DataFrame
    fpath: file path, if extension is provided it will overwrite formats
    otherwise formats is considered
    ext: str or list of string with file extensions
    index: bool add index in exporting (typically True for results, False for data)
    """

    if not (isinstance(x, _pd.DataFrame) or isinstance(x, dict)):
        msg = "x deve essere un pd.DataFrame o un dict di pd.DataFrame"
        raise ValueError(msg)

    if isinstance(ext, str): #uniforma
        ext = [ext]

    path = _Path(path)
    path_has_suffix = path.suffix != ""
    if path_has_suffix:
        used_formats = [path.suffix.replace(".", "")]
    else:
        used_formats = ext

    if "xlsx" in used_formats:
        xlsx_path = path if path_has_suffix else path.with_suffix(".xlsx")
        with _pd.ExcelWriter(xlsx_path) as writer:
            if isinstance(x, _pd.DataFrame):
                x.to_excel(writer, sheet_name="Foglio 1", index=index)
            elif isinstance(x, dict):
                for k, v in x.items():
                    # preprocess the key of the dict, it must be an accepted excel
                    # sheetname. Trim it to the first x characters
                    k = _fix_varnames(k)[:31]
                    v.to_excel(writer, sheet_name=k, index=index)

    if "csv" in used_formats:
        if isinstance(x, _pd.DataFrame):
            x.to_csv(path if path_has_suffix else path.with_suffix(".csv"), index=index)
        elif isinstance(x, dict):
            # use dict key as postfix
            for k, v in x.items():
                csv_path = path.parent / (str(path.stem) + f"_{k}.csv")
                v.to_csv(csv_path, index=index)

    if "pkl" in used_formats:
        if isinstance(x, _pd.DataFrame):
            x.to_pickle(path if path_has_suffix else path.with_suffix(".pkl"))
        elif isinstance(x, dict):
            # use dict key as postfix
            for k, v in x.items():
                pkl_path = path.parent / (str(path.stem) + f"_{k}.pkl")
                v.to_pickle(pkl_path)

    if "R" in used_formats:
        if isinstance(x, _pd.DataFrame):
            _rdf(x, path if path_has_suffix else path.with_suffix(".R"))
        elif isinstance(x, dict):
            # use dict key as postfix
            for k, v in x.items():
                R_path = path.parent / (str(path.stem) + f"_{k}.R")
                _rdf(v, R_path)



# ------------------------------------
# LaTeX stuff
# ------------------------------------

# Thanks PyLaTeX guys
_latex_special_chars = {
    '&': r'\&',
    '%': r'\%',
    '$': r'\$',
    '#': r'\#',
    '_': r'\_',
    '{': r'\{',
    '}': r'\}',
    '~': r'\textasciitilde{}',
    '^': r'\^{}',
    '\\': r'\textbackslash{}',
    '\n': '\\newline%\n',
    '-': r'{-}',
    '\xA0': '~',  # Non-breaking space
    '[': r'{[}',
    ']': r'{]}',
}


def _latex_escape(s):
    """
    latex_escape from PyLaTeX

    s: a str or something coercible to it
    >>> print(latex_escape("asd_foo_bar"))
    """
    string = str(s)
    return ''.join(_latex_special_chars.get(c, c) for c in string)


def latex_table(
    tab,
    label: str = "",
    caption: str = "",
    position: str | None = None,
    float_format: str = "%.1f",
    column_format: str | None = None,
):
    """Print a table (a pd.DataFrame or inheriting object with
    to_latex method) with sensible defaults.

    tab: a pd.DataFrame or other object with .to_latex method
    label (str): posta dopo "fig:" in LaTeX
    caption (str): caption tabella LaTeX
    position (str): lettere posizionamento tabella (ad esempio https://stackoverflow.com/questions/1673942)
    """
    if (label == "") or (not isinstance(label, str)):
        msg = "Please provide a label for the table."
        raise ValueError(msg)
    caption = (
        label.capitalize().replace("_", " ") if caption == "" else caption
    )
    latex_caption = _latex_escape(caption)
    latex_label = 'tab:' + label
    if isinstance(tab, _pd.DataFrame):
        tab.columns.name = None
    if (column_format == None) and isinstance(tab, _pd.DataFrame):
        ncols = tab.shape[1]
        column_format = "".join(["l"] + ["r"] * ncols)
    # per avere il centering è necessario impostare lo stile
    # https://pandas.pydata.org/docs/reference/api/pandas.io.formats.style.Styler.to_latex.html
    content = tab.to_latex(
        # fissi
        na_rep="",
        index=True,
        index_names=False,
        escape=True,
        # position_float='centering', # one day. maybe.
        # variabili
        label=latex_label,
        caption=latex_caption,
        position=position,
        float_format=float_format,
        column_format=column_format,
    )
    print(content)



def export_tables(tabs_dict):
    """Latex print and excel Export a dict of tables (caption as key)

    Example usage:
    --------------
    exported = {
       "Descrittive": desc_df,
       "Analisi 1"  : analysis1_df,
       "Analisi 2"  : analysis2_df
    }
    export_tables(exported)    
    """
    # excel
    export_data(tabs_dict, "outputs/tables.xlsx")
    # latex
    labels = []
    for caption, tab in tabs_dict.items():
        lab = _fix_varnames(caption)
        labels.append(lab)
        latex_table(tab, label = lab, caption = caption)
    # return latex references
    refs = ["\ref{tab:" + lab + "}" for lab in labels]
    return refs
